#!/usr/bin/env python
# coding: utf-8
import time
import os
import sys
import logging
import threading
import logging.handlers
from datetime import datetime
from glob import glob

from intelliw.utils import string_to_function, gen_random_str
from intelliw.utils.colorlog import ColoredFormatter
from intelliw.utils.context import header_ctx

framework_logger = None
user_logger = None


def log_path():
    path = os.environ.get("INTELLIW_LOG_PATH") or \
           os.environ.get("intelliw.logger.path") or \
           "./logs/"
    code = os.environ.get("INSTANCE_POD_NAME", "pod")
    if not sys.platform.startswith('win') and os.environ.get('isPremises', 'false') != 'true':
        path = os.path.join(path, code, f'{datetime.now().strftime("%Y-%m-%d:%H:%M:%S")}')
    if not os.path.exists(path):
        os.makedirs(path)
    return path


class LogCfg:
    level = os.environ.get("intelliw.logger.level", logging.INFO)

    format = ('[%(traceid)s] [PID%(process)d] -%(name)s- %(asctime)s '
              '| %(levelname)4s | %(filename)s:%(lineno)s: %(message)s')

    colorful_format = ('[%(traceid)s] %(log_color)s[PID%(process)d] -%(name)s- %(asctime)s '
                       '| %(levelname)4s | %(filename)s:%(lineno)s: %(message)s')

    data_format = '%Y-%m-%d %H:%M:%S'
    path = log_path()


def inject_code():
    class MyLogRecord(logging.LogRecord):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.traceid = '0'

    logging.setLogRecordFactory(MyLogRecord)
    logging.PercentStyle._format = string_to_function('''
def _format(self, record):
    if '%(traceid)' not in self._fmt:
        self._fmt = f'[%(traceid)s] {self._fmt}'
    if not hasattr(record, 'traceid'):
        setattr(record, 'traceid', '0')
    return self._fmt % record.__dict__
    ''', "_format")


inject_code()


def is_file_handler(handler):
    return isinstance(handler, logging.FileHandler)


def update_file_handlers_level(new_level):
    # user logger
    logger_dict = logging.root.manager.loggerDict
    for logger_name, logger_obj in logger_dict.items():
        if isinstance(logger_obj, logging.Logger):
            for handler in logger_obj.handlers:
                if is_file_handler(handler):  # 检查是否为文件写入 handler
                    handler.setLevel(new_level)
                    logger_obj.setLevel(new_level)  # 同步更新 logger 的 level
    # root
    for handler in logging.root.handlers:
        if is_file_handler(handler):
            handler.setLevel(new_level)


def inject_level(level):
    level = logging._checkLevel(level)

    def getEffectiveLevel(self):
        """
        Get the effective level for this logger.

        Loop through this logger and its parents in the logger hierarchy,
        looking for a non-zero logging level. Return the first one found.
        """
        return level

    logging.Logger.getEffectiveLevel = getEffectiveLevel
    logging.Logger.manager._clear_cache()
    update_file_handlers_level(level)


class TraceIDFilter(logging.Filter):
    def filter(self, record):
        header = header_ctx.get()
        if isinstance(header, dict):
            record.traceid = header.get('traceId', '0')
        else:
            record.traceid = '0'
        return True


class CustomRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, base_filename, *args, **kwargs):
        self.base_filename = base_filename
        super().__init__(base_filename, *args, **kwargs)

    def doRollover(self):
        # 如果日志文件存在且需要滚动
        if self.stream:
            self.stream.close()
            self.stream = None

        # 获取当前时间作为文件名的一部分
        current_time = time.strftime("%Y-%m-%d_%H-%M-%S")
        rollover_filename = f"{self.base_filename}_{current_time}.log"

        # 重命名当前日志文件
        if os.path.exists(self.base_filename):
            os.rename(self.base_filename, rollover_filename)

        # 调用父类的方法来处理日志文件切换
        self.mode = 'w'
        self.stream = self._open()

        # 检查日志文件数量，删除最旧的多余文件
        self.clean_old_logs()

    def clean_old_logs(self):
        # 获取当前目录下所有的滚动日志文件
        log_files = glob(f"{self.base_filename}_*.log")
        if len(log_files) > self.backupCount:
            # 按文件的创建时间排序
            log_files.sort(key=os.path.getctime)
            # 删除多余的旧日志文件
            for log_file in log_files[:len(log_files) - self.backupCount]:
                os.remove(log_file)


class Logger:
    _instance_lock = threading.Lock()

    def __new__(cls):
        """ 单例,防止调用生成更多 """
        if not hasattr(Logger, "_instance"):
            with Logger._instance_lock:
                if not hasattr(Logger, "_instance"):
                    if not os.path.exists(LogCfg.path):
                        os.makedirs(LogCfg.path)
                    Logger._instance = object.__new__(cls)
                    Logger.__global_logger(cls)
        return Logger._instance

    def __global_logger(cls):
        root_logger = logging.getLogger()
        root_logger.addFilter(TraceIDFilter())

        # set up the logger to write into file
        if os.access(LogCfg.path, os.W_OK):
            time_file_handler = CustomRotatingFileHandler(
                os.path.join(LogCfg.path, f'iw-algo-fx.log'),
                maxBytes=1024 * 1024 * 100,
                backupCount=50)
            time_file_handler.setLevel(LogCfg.level)
            time_file_handler.setFormatter(
                logging.Formatter(LogCfg.format)
            )
            root_logger.addHandler(time_file_handler)

        # Setup the logger to write into stdout
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(ColoredFormatter(LogCfg.colorful_format))
        root_logger.addHandler(consoleHandler)

        logging.Logger.root = root_logger
        logging.root = logging.Logger.root
        logging.Logger.manager = logging.Manager(logging.Logger.root)
        logging.root.setLevel(LogCfg.level)
        logging.LogRecord.traceid = '0'

        inject_level(LogCfg.level)
        logging.info("root logger init, level:%s", logging.getLevelName(LogCfg.level))

    def __init__(self):
        self.framework_logger = self._get_logger(
            "Framework Log", level=LogCfg.level, filename="iw-algo-fx-framework.log")
        self.user_logger = self._get_logger(
            "Algorithm Log", level=LogCfg.level, filename="iw-algo-fx-user.log")

    @staticmethod
    def _get_logger(logger_type, level=logging.INFO, format=None, filename=None):
        logger = logging.getLogger(logger_type)
        logger.addFilter(TraceIDFilter())

        if logger.handlers:
            return logger

        format = format or LogCfg.colorful_format
        if filename is not None:
            if os.access(LogCfg.path, os.W_OK):
                time_file_handler = CustomRotatingFileHandler(
                    os.path.join(LogCfg.path, filename),
                    maxBytes=1024 * 1024 * 50,
                    backupCount=20)
                formatter = ColoredFormatter(
                    format, datefmt=LogCfg.data_format)
                time_file_handler.setLevel(level)
                time_file_handler.setFormatter(formatter)
                logger.addHandler(time_file_handler)
        return logger


def _get_framework_logger():
    global framework_logger
    if framework_logger is None:
        framework_logger = Logger().framework_logger
    return framework_logger


def _get_algorithm_logger():
    global user_logger
    if user_logger is None:
        user_logger = Logger().user_logger
    return user_logger


def get_logger(name: str = "user", level: str = "INFO", format: str = None, filename: str = None):
    """get custom logs

    Args:
        name (str, optional): Logger unique name. Defaults to "user".
        level (str, optional): Logger level. Defaults to "INFO".
        format (str, optional): Format the specified record. Defaults to None.
        filename (str, optional): Save the name of the log file. Defaults to None.

    Returns:
        logger
    """
    return Logger()._get_logger(name, level, format, filename)



def select_columns(data, cfg=None):
    return_meta = []
    return_result = []

    column_list = cfg['columns_name'].strip().split(',')

    # form code_list
    code_list = []
    for tmp_dict in data['meta']:
        code_list.append(tmp_dict['code'])
    # the index list of selected column in data
    selected_column_index_list = []
    for code in column_list:
        try:
            idx = code_list.index(code)
        except ValueError:
            raise ValueError(f'Dataset not contain column: {code}')
        selected_column_index_list.append(idx)
        return_meta.append(data['meta'][idx])
    for i in range(len(data['result'])):
        tmp_list = []
        for j in selected_column_index_list:
            tmp_list.append(data['result'][i][j])
        return_result.append(tmp_list)

    data['meta'] = return_meta
    data['result'] = return_result

    return data



'''
Author: Hexu
Date: 2022-04-25 15:16:48
LastEditors: Hexu
LastEditTime: 2023-05-11 09:45:28
FilePath: /iw-algo-fx/intelliw/datasets/datasource_iwimgdata.py
Description: 图片数据集
'''
import json
import os
import shutil

from intelliw.config import config
from intelliw.datasets.datasource_base import AbstractDataSource, DataSourceReaderException, DatasetSelector, \
    DataSourceType
from intelliw.utils.exception import DataSourceDownloadException
from intelliw.utils import iuap_request
from intelliw.utils.logger import _get_framework_logger
from concurrent.futures import ThreadPoolExecutor, as_completed

logger = _get_framework_logger()


def is_classification():
    """Checks whether the dataset is for image classification.

    Returns:
        bool: True if the dataset is for image classification, False otherwise.
    """
    from intelliw.utils.global_val import gl
    return gl.model_type == 12


class CocoProcess:
    coco_type = 3
    coco_config = None
    licenses = None
    info = None
    categories = None
    set_config = None
    current_no = 0

    @classmethod
    def set_coco_info(cls, instance, prefix):
        cls.info = instance['info']
        cls.licenses = instance['licenses']
        cls.categories = instance['categories']
        cls.coco_config = cls.__gen_config(
            instance['images'], instance['annotations'], prefix)
        cls.reset_config()

    @classmethod
    def reset_config(cls):
        cls.set_config = {'licenses': cls.licenses, 'info': cls.info,
                          'categories': cls.categories, 'images': [], 'annotations': []}

    @classmethod
    def __gen_config(cls, images, annotations, prefix):
        a_map, i_config = {}, {}
        for a in annotations:
            image_id = a['image_id']
            a_map[image_id] = a_map.get(image_id, [])
            a_map[image_id].append(a)
        for i in images:
            i['file_name'] = prefix + i['file_name']
            i_config[i['file_name']] = {
                'image': i, 'annotation': a_map[i['id']]}
        return i_config

    @classmethod
    def gen_config(cls, filename):
        meta = cls.coco_config.get(filename)
        if meta is None:
            return f"image:{filename} annotation not exist"
        cls.set_config['images'].append(meta['image'])
        cls.set_config['annotations'].extend(meta['annotation'])
        return None

    @classmethod
    def flush(cls, path, dataset_no):
        if dataset_no > 1 and os.path.exists(path):
            with open(path, "r", encoding="utf-8") as rfp:
                cfg = json.load(rfp)
                cls.set_config['images'].extend(cfg['images'])
                cls.set_config['annotations'].extend(cfg['annotations'])
        with open(path, 'w') as wfp:
            json.dump(cls.set_config, wfp, ensure_ascii=False)


class DataSourceIwImgData(AbstractDataSource):
    """
    非结构化存储数据源
    图片数据源
    """

    # 共几个数据集， 当大于1时，写到同一个文件
    dataset_count = 0

    def __init__(self, input_address, get_row_address, ds_id, ds_type, tenant_id=None):
        """
        智能分析数据源
        :param input_address:   获取数据 url
        :param get_row_address: 获取数据总条数 url
        :param ds_id:   数据集Id
        :param ds_type: cv标注类型: 0-自有 1-labelme 2-voc 3-coco
        :param is_classification 是否为分类数据集，需要做样本均分
        """
        DataSourceIwImgData.dataset_count += 1
        self.dataset_no = DataSourceIwImgData.dataset_count

        self.input_address = input_address
        self.get_row_address = get_row_address
        self.ds_id = ds_id
        self.ds_type = ds_type
        self.__total = None
        self.__gen_img_dir()
        self.is_classification = is_classification()
        self.tenant_id = tenant_id if tenant_id is not None else config.TENANT_ID

    def __gen_img_dir(self):
        logger = _get_framework_logger()
        filepath = os.path.join('./', config.CV_IMG_FILEPATH)
        if os.path.exists(filepath) and self.dataset_no == 1:
            logger.warn(f"图片数据保存路径存在:{filepath}, 正在删除路径内容")
            shutil.rmtree(filepath, ignore_errors=True)
        os.makedirs(config.CV_IMG_ANNOTATION_FILEPATH, 0o755, True)
        os.makedirs(config.CV_IMG_TRAIN_FILEPATH, 0o755, True)
        os.makedirs(config.CV_IMG_VAL_FILEPATH, 0o755, True)
        os.makedirs(config.CV_IMG_TEST_FILEPATH, 0o755, True)

    def total(self):
        if self.__total is not None:
            return self.__total
        params = {'dsId': self.ds_id, 'yTenantId': self.tenant_id}
        response = iuap_request.get(self.get_row_address, params=params, timeout=config.DATA_SOURCE_READ_TIMEOUT)
        if 200 != response.status:
            msg = "获取行数失败，url: {}, response: {}".format(
                self.get_row_address, response)
            raise DataSourceReaderException(msg)

        row_data = response.json
        self.__total = row_data['data']

        if not isinstance(self.__total, int):
            msg = "获取行数返回结果错误, response: {}, request_url: {}".format(
                row_data, self.get_row_address)
            raise DataSourceReaderException(msg)
        return self.__total

    def reader(self, page_size=10000, offset=0, limit=0, transform_function=None, dataset_type='train_set'):
        r = self.__Reader(self.input_address, self.ds_id, self.ds_type, self.total(), dataset_type, self.tenant_id,
                          page_size, transform_function, is_classification=self.is_classification)
        r.dataset_no = self.dataset_no
        return r

    def download_images(self, images, transform_function=None, dataset_type='train_set'):
        r = self.reader(transform_function=transform_function,
                        dataset_type=dataset_type)
        r.dataset_no = self.dataset_no
        r.set_download(images)
        return r()

    class __Reader:
        def __init__(self, input_address, ds_id, ds_type, total, process_type, tenant_id, page_size=100,
                     transform_function=None, is_classification=False):
            self.input_address = input_address
            self.ds_id = ds_id
            self.ds_type = ds_type
            self.total_data = total
            self.total_read = 0
            self.page_size = max(100, page_size)
            self.page_num = 0
            self.process_type = process_type
            self.transform_function = transform_function
            self.worker = min(3, max(1, os.cpu_count()))
            self.dataset_no = 0
            self.is_classification = is_classification
            self.tenant_id = tenant_id

        def get_data_bar(self):
            """数据拉取进度条"""
            try:
                proportion = round(
                    (self.total_read / self.total_data) * 100, 2)
                logger.info(
                    f"图片数据集[{self.dataset_no}]元数据下载获取中: 共{self.total_data}条数据, 已获取{self.total_read}条, 进度{proportion}%")
            except:
                pass

        @property
        def iterable(self):
            return True

        def __iter__(self):
            return self

        def __next__(self):
            logger = _get_framework_logger()
            if self.total_read >= self.total_data:
                logger.info(
                    f'图片数据集[{self.dataset_no}]元数据下载完成，准备下载图片，共读取原始数据 {self.total_read} 条')
                raise StopIteration

            self.get_data_bar()

            try:
                with ThreadPoolExecutor(max_workers=self.worker) as executor:
                    futures, data_result = [], []
                    for i in range(self.worker * 2):
                        futures.append(executor.submit(
                            self._read_page, self.page_num, self.page_size
                        ))
                        self.page_num += 1

                    for f in futures:
                        page = f.result()
                        data = page['data']['content']
                        data_result.extend(self._label_process(data))

                    self.total_read += len(data_result)
                    if len(data_result) == 0:
                        logger.info(
                            f'图片数据集[{self.dataset_no}]元数据下载完成，准备下载图片，共读取原始数据 {self.total_read} 条')
                        raise StopIteration
                    return data_result
            except StopIteration:
                raise StopIteration
            except Exception as e:
                logger.exception(
                    f"图片数据集[{self.dataset_no}]数据源读取失败, input_address: [{self.ds_id}]")
                raise DataSourceReaderException(f'图片数据源读取失败:{e}')

        def _label_process(self, data):
            if self.is_classification:
                for idx, d in enumerate(data):
                    labels = d.get('labels', [])
                    if len(labels) == 0:
                        raise Exception("分类数据集含有未标注的图片")
                    data[idx]['labels'] = labels[0]['name']
            return data

        def _read_page(self, page_index, page_size):
            """
            图片数据接口，分页读取数据
            :param page_index: 页码，从 0 开始
            :param page_size:  每页大小
            :return:
            """
            request_data = {'dsId': self.ds_id, 'pageNumber': page_index,
                            'pageSize': page_size,
                            'yTenantId': self.tenant_id,
                            'type': self.ds_type}
            response = iuap_request.get(
                url=self.input_address, params=request_data, timeout=config.DATA_SOURCE_READ_TIMEOUT)
            response.raise_for_status()
            return response.json

        def set_download(self, page):
            logger = _get_framework_logger()
            prefix = str(
                self.dataset_no) if DataSourceIwImgData.dataset_count > 1 else ""
            logger.info(f"图片数据集[{self.dataset_no}]开始下载")
            if self.ds_type == CocoProcess.coco_type:
                CocoProcess.reset_config()
                if CocoProcess.current_no != self.dataset_no:
                    CocoProcess.current_no = self.dataset_no
                    annotation_urls = page[0]['annotationUrl']
                    annotation = save_file(
                        annotation_urls, timeout=90, is_img=False)
                    if annotation is None:
                        raise DataSourceReaderException(
                            f'图片数据级标注信息有误，标注文件地址：{annotation_urls}')
                    annotation = json.loads(annotation)
                    CocoProcess.set_coco_info(annotation, prefix)

            err_count, success_count, err_msg = 0, 0, ""
            with ThreadPoolExecutor(max_workers=self.worker) as executor:
                futures = [executor.submit(
                    self.__download, p, prefix) for p in page]
                for f in as_completed(futures):
                    try:
                        f.result()
                        success_count += 1
                        if success_count % 10 == 0:
                            logger.info(
                                f"图片数据集[{self.dataset_no}]下载中: {success_count}/{self.total_data}")
                    except Exception as e:
                        err_count += 1
                        err_msg = e
                        logger.error(
                            f'图片数据集[{self.dataset_no}]下载错误: {e}, total {err_count}')

            if err_count > self.total_data * 1e-2 and success_count / err_count < 1000:
                raise DataSourceDownloadException(
                    f"图片数据集[{self.dataset_no}]下载失败的图片过多,  successed: {success_count},  failed: {err_count}, msg: {err_msg}")

            if self.ds_type == CocoProcess.coco_type:
                filename = self.process_type + ".json"
                CocoProcess.flush(os.path.join(
                    '.', config.CV_IMG_ANNOTATION_FILEPATH, filename), self.dataset_no)

            logger.info(
                f"图片数据集[{self.dataset_no}] {self.process_type} 下载完成， 成功{success_count}, 失败{err_count}")

        def __download(self, page, prefix):
            url = page['url']
            annotation_url = page['annotationUrl']
            filename = prefix + page['rowFileName']
            annotationname = filename.replace(
                page['fileNameType'], page['annotationType'])

            # 图片下载， 图片可能伴随特征工程
            if self.process_type == 'train_set':
                process_file = config.CV_IMG_TRAIN_FILEPATH
            elif self.process_type == 'validation_set':
                process_file = config.CV_IMG_VAL_FILEPATH
            else:
                process_file = config.CV_IMG_TEST_FILEPATH

            filepath = os.path.join('.', process_file, filename)
            if save_file(url, filepath, self.transform_function, timeout=config.DATA_SOURCE_READ_TIMEOUT, is_img=True) is None:
                raise Exception(f"image:{filename} download error")

            # 标注下载或写入内存
            if self.ds_type == CocoProcess.coco_type:
                err = CocoProcess.gen_config(filename)
                if err is not None:
                    raise KeyError(f"CocoProcess.gen_config error: {err}")
            else:
                filepath = os.path.join(
                    '.', config.CV_IMG_ANNOTATION_FILEPATH, annotationname)
                if save_file(annotation_url, filepath, timeout=config.DATA_SOURCE_READ_TIMEOUT, is_img=False) is None:
                    raise Exception(
                        f"annotation:{annotationname} download error")

            return filename

        def __call__(self):
            abspath = os.path.abspath('.')
            return {'path': os.path.join(abspath, config.CV_IMG_FILEPATH),
                    'train_set': os.path.join(abspath, config.CV_IMG_TRAIN_FILEPATH),
                    'val_set': os.path.join(abspath, config.CV_IMG_VAL_FILEPATH),
                    'test_set': os.path.join(abspath, config.CV_IMG_TEST_FILEPATH),
                    'annotations': os.path.join(abspath, config.CV_IMG_ANNOTATION_FILEPATH)}


def save_file(url, filepath=None, transform_function=None, timeout=config.DATA_SOURCE_READ_TIMEOUT, is_img=True):
    logger = _get_framework_logger()
    try:
        response = iuap_request.download(url=url, timeout=timeout)
        status = response.status
        if status == 200:
            data = response.body
            if is_img and transform_function:
                data = transform_function(data)
            if filepath:
                with open(filepath, 'wb') as fp:
                    fp.write(data)
                return filepath
            return data
        logger.error("http get url %s failed, status is %s", url, status)
        return None
    except Exception as e:
        logger.error("http get url %s failed, error is %s", url, e)
        return None


DatasetSelector.register_func(
    DataSourceType.IW_IMAGE_DATA, DataSourceIwImgData, {
        "input_address": "INPUT_ADDR",
        "get_row_address": "INPUT_GETROW_ADDR",
        "ds_id": "INPUT_DATA_SOURCE_ID",
        "ds_type": "INPUT_DATA_SOURCE_TRAIN_TYPE",
        "tenant_id": "TENANT_ID"})

if __name__ == '__main__':
    save_file(
        url='https://intelliw-console.oss-cn-beijing-internal.aliyuncs.com/AI-CONSOLE/iwdata/0000L4PB5VYW84IL8O0000/1826085601888698376/voc/2b820359dbbd83c981b6357e6edc4bee.xml?Expires=3272593373&OSSAccessKeyId=LTAI5tR6jdSeCNyRfTZ5CN2s&Signature=8E7KGaKrBIEgnAaBpLX4bBRLNok%3D',
        filepath="./a.xml", is_img=False)



import os
import yaml
from intelliw.config import config

key = "iuapaipaas_iwinfer210706".encode('utf-8')
iv = b'iuapaipaas_iwinf'


class AttrDict(dict):
    """Single level attribute dict, NOT recursive"""

    def __init__(self, **kwargs):
        super(AttrDict, self).__init__()
        super(AttrDict, self).update(kwargs)

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError("object has no attribute '{}'".format(key))


global_config = AttrDict()

READER_KEY = '_READER_'


def load_config(file_path):
    """
    Load config from file.

    Args:
        file_path (str): Path of the config file to be loaded.

    Returns: global config
    """
    _, ext = os.path.splitext(file_path)
    assert ext in ['.yml', '.yaml'], "only support yaml files for now"

    cfg = AttrDict()
    if config.IS_SPECIALIZATION == 2:
        with open(file_path, 'rb') as f:
            text = f.read()
            model = aes_cbc_decrypt(text)
            cfg = merge_config(yaml.load(model, Loader=yaml.Loader), cfg)
    else:
        with open(file_path, 'r', encoding='UTF-8') as f:
            cfg = merge_config(yaml.load(f, Loader=yaml.Loader), cfg)

    if READER_KEY in cfg:
        reader_cfg = cfg[READER_KEY]
        if reader_cfg.startswith("~"):
            reader_cfg = os.path.expanduser(reader_cfg)
        if not reader_cfg.startswith('/'):
            reader_cfg = os.path.join(os.path.dirname(file_path), reader_cfg)

        with open(reader_cfg) as f:
            merge_config(yaml.load(f, Loader=yaml.Loader))
        del cfg[READER_KEY]

    merge_config(cfg)
    return global_config


def load_single_config(file_path):
    _, ext = os.path.splitext(file_path)
    assert ext in ['.yml', '.yaml'], "only support yaml files for now"

    if config.IS_SPECIALIZATION == 2:
        with open(file_path, 'rb') as f:
            text = f.read()
            model = aes_cbc_decrypt(text)
            cfg = yaml.load(model, Loader=yaml.Loader)
    else:
        with open(file_path, 'r', encoding='UTF-8') as f:
            cfg = yaml.load(f, Loader=yaml.Loader)
    return cfg


def dump_config(data, file_path):
    _, ext = os.path.splitext(file_path)
    assert ext in ['.yml', '.yaml'], "only support yaml files for now"

    if config.IS_SPECIALIZATION == 2:
        with open(file_path, 'wb') as f:
            model = yaml.dump(data, default_flow_style=False)
            f.write(model)
    else:
        with open(file_path, 'w', encoding='UTF-8') as f:
            model = yaml.dump(data, default_flow_style=False,
                              allow_unicode=True)
            f.write(model)


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    Args:
        dct: dict onto which the merge is executed
        merge_dct: dct merged into dct

    Returns: dct
    """
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and
                isinstance(merge_dct[k], dict)):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]
    return dct


def merge_config(config, another_cfg=None):
    """
    Merge config into global config or another_cfg.

    Args:
        config (dict): Config to be merged.

    Returns: global config
    """
    global global_config
    dct = another_cfg if another_cfg is not None else global_config
    return dict_merge(dct, config)


def aes_cbc_decrypt(text):
    from Crypto.Cipher import AES
    cryptos = AES.new(key, AES.MODE_CBC, iv)
    plain_text = cryptos.decrypt(text)
    # remove ETX (end of text)
    a = bytes.decode(plain_text)
    return remove_ctl(a)


def remove_ctl(input_str):
    result = ''
    for c in input_str:
        ascii_val = ord(c)
        valid_chr_list = [
            9,  # 9=\t=tab
            10,  # 10=\n=LF=Line Feed=换行
            13,  # 13=\r=CR=回车
        ]
        # filter out others ASCII control character, and DEL=delete
        valid_chr = True
        if ascii_val == 0x7F:
            valid_chr = False
        elif (ascii_val < 32) and (ascii_val not in valid_chr_list):
            valid_chr = False

        if valid_chr:
            result += c

    return result



'''
Author: Hexu
Date: 2022-07-25 10:36:04
LastEditors: Hexu
LastEditTime: 2023-03-21 10:08:46
FilePath: /iw-algo-fx/intelliw/feature.py
Description: 统一功能包入口
'''
# 🙅🙅🙅不要删除🙅🙅🙅
from intelliw.interface.iwapi.iwapp import (
    Application,
    set_initializer
)
from intelliw.core.linkserver import linkserver
from intelliw.utils.logger import _get_algorithm_logger as get_logger, _get_framework_logger
from intelliw.utils import exception
from intelliw.utils.storage_service import FileTransferDevice
from intelliw.datasets.datasets import get_datasource_writer as OutPutWriter
from intelliw.utils import context

try:
    from intelliw.utils.intelliwapi.workers import IntelliwWorker
except Exception as e:
    _get_framework_logger().warning(
        "\033[33mGunicorn can not used\033[0m, error: %s", e)

try:
    from intelliw.utils.spark_process.spark import Spark
except ImportError as e:
    _get_framework_logger().warning(
        "\033[33mIf want use spark, you need: pip install pyspark\033[0m, error: %s", e)




import logging
import os
import uvicorn
from intelliw.utils.colorlog import ColoredFormatter
from intelliw.utils.logger import _get_framework_logger, TraceIDFilter


def default_config(bind, workers=None):
    config = {
        'bind': bind,
        'accesslog': '-', 'errorlog': '-',
        'timeout': 6000,
        'workers': workers or 1,
        'worker_class': 'intelliw.feature.IntelliwWorker',
        'logger_class': CustomLogger
    }
    return config


try:
    import gunicorn.app.base
    from gunicorn import glogging


    class CustomLogger(glogging.Logger):
        """Custom logger for Gunicorn log messages."""

        def __set_handler(self, logger, formatter, handler):
            h = self._get_gunicorn_handler(logger)
            if h:
                logger.handlers.remove(h)
            h.setFormatter(formatter)
            h._gunicorn = True
            logger.addHandler(h)
            logger.addHandler(handler)

        def setup(self, cfg):
            """Configure Gunicorn application logging configuration."""
            super().setup(cfg)

            format_string = '%(log_color)s[%(process)d] -System Log-  %(asctime)s | %(levelname)4s | %(message)4s'
            formatter = ColoredFormatter(format_string)

            framework_handler = _get_framework_logger().handlers[0]

            # Override Gunicorn's `error_log` configuration.
            self.__set_handler(self.error_log, formatter, framework_handler)
            self.__set_handler(self.access_log, formatter, framework_handler)


    class GunServer(gunicorn.app.base.BaseApplication):

        def __init__(self, app, options=None, logger=None, limit_concurrency=100):
            self.options = options or {}
            self.application = app
            self.logger = logger

            super().__init__()
            # self.cfg.set('max_requests', limit_concurrency)

        def load_config(self):
            config = {key: value for key, value in self.options.items()
                      if key in self.cfg.settings and value is not None}
            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.application

        def run(self):
            super().run()
except Exception as e:
    _get_framework_logger().warning('Windows can not use Gunicorn, Uvicorn service import')


class UvicornServer:

    def __init__(self, app, host='0.0.0.0', port=8888, workers=None, limit_concurrency=9999):
        config = uvicorn.Config(
            app,
            host=host,
            port=int(port),
            workers=workers,
            limit_concurrency=limit_concurrency
        )
        self.server = uvicorn.Server(config)

    def run(self):
        class NoMsgFilter(logging.Filter):
            def filter(self, record):
                return record.getMessage().find('/healthcheck') < 0

        logger = logging.getLogger("uvicorn.error")
        logger.handlers = logging.getLogger().handlers
        logger.addFilter(TraceIDFilter())
        logger.addFilter(NoMsgFilter())
        logger.propagate = False

        logger = logging.getLogger("uvicorn.access")
        logger.handlers = logging.getLogger().handlers
        logger.addFilter(TraceIDFilter())
        logger.addFilter(NoMsgFilter())
        logger.propagate = False

        self.server.run()


if __name__ == '__main__':
    # options = {
    #     'bind': '%s:%s' % ('127.0.0.1', '8080'),
    #     'workers': number_of_workers(),
    # }
    # StandaloneApplication(handler_app, options).run()
    pass



'''
Author: hexu
Date: 2021-10-14 14:54:05
LastEditTime: 2023-05-25 15:24:05
LastEditors: Hexu
Description: 数据集切分工具
FilePath: /iw-algo-fx/intelliw/datasets/spliter.py
'''
from abc import ABCMeta, abstractmethod
import math
from random import Random
from typing import Iterable
from intelliw.utils.exception import DatasetException
from intelliw.config import config
from intelliw.utils.global_val import gl
from intelliw.utils.logger import _get_framework_logger
from intelliw.datasets.datasource_base import DataSourceType


class SpliterClass:
    nothing = -1
    sequential = 0
    random = 1
    random_by_target = 2


def get_set_spliter(data, source_type, not_spliter=False):
    logger = _get_framework_logger()
    trr = config.TRAIN_DATASET_RATIO
    var = config.VALID_DATASET_RATIO
    ter = config.TEST_DATASET_RATIO
    msg = "\033[33mDataset Split Reader: %s\033[0m"

    if trr == 1 and (var + ter) == 0:
        not_spliter = True
    logger.info("SplitRatio: trr:%f,var:%f,ter:%f", trr, var, ter)
    if not data or not_spliter or \
            config.DATA_SPLIT_MODE == SpliterClass.nothing:
        logger.info(msg, "no Spliter")
        return NoSpliter(data, trr, var, ter, source_type)
    elif config.DATA_SPLIT_MODE == SpliterClass.random:
        target_col, err = target_verify()
        if err is None:  # 分类算法根据目标列特殊处理
            logger.info(msg, "Random Spliter by Target")
            return TargetRandomSpliter(data, trr, var, ter, target_col, source_type)
        else:
            logger.info(msg, "Random Spliter")
            return ShuffleSpliter(data, trr, var, ter, source_type)
    elif config.DATA_SPLIT_MODE == SpliterClass.random_by_target:
        target_col, err = target_verify()
        if err is not None:
            raise DatasetException(err)
        logger.info(msg, "Random Spliter by Target")
        return TargetRandomSpliter(data, trr, var, ter, target_col, source_type)
    else:  # 不指定都用顺序
        logger.info(msg, "Sequential Spliter")
        return SequentialSpliter(data, trr, var, ter, source_type)


def target_verify():
    target_col, err = None, None

    # 图像分类
    if gl.model_type == 12:
        return "labels", err

    # 表格分类
    if not gl.model_type == 3:
        err = "只有分类算法才能使用该规则"
        return target_col, err

    target_metadata = gl.target_metadata
    if target_metadata is None or len(target_metadata) == 0:
        err = "配置文件(algorithm.yaml)中未设置target相关数据"
        return target_col, err

    if len(target_metadata) > 1:
        err = "目前只支持针对单目标列的数据shuffle处理"
        return target_col, err

    # 先从特征工程后的target中获取
    # 无特征工程修改后的目标列， 再从源数据中获取
    target_cols = gl.get("target_cols", [])
    if len(target_cols) > 0:
        target_col = target_cols[0]["col"]
    else:
        target_col = target_metadata[0]["target_col"]
    if type(target_col) != int:
        err = f"目标列下标类型错误:targetCol类型应为int, 当前数据: {target_col}-{type(target_col)}"
        return target_col, err

    return target_col, err


def check_split_ratio(train_ratio, valid_ratio, test_ratio):
    assert 0 < train_ratio <= 1, f"数据集比例不正确, 训练集比例{train_ratio}"
    assert 0 <= valid_ratio < 1, f"数据集比例不正确, 验证集比例{valid_ratio}"
    assert 0 <= test_ratio < 1, f"数据集比例不正确, 测试集比例{test_ratio}"
    assert train_ratio + valid_ratio + test_ratio < 1.09, "数据集比例不正确, 总和不为1"


def get_set_count(length, train_ratio, valid_ratio, test_ratio):
    train_num = math.floor(length * float(train_ratio))
    if test_ratio == 0:
        valid_num = length - train_num
        test_num = 0
    else:
        valid_num = math.floor(length * float(valid_ratio))
        test_num = max(length - train_num - valid_num, 0)
    return train_num, valid_num, test_num


class DataSetSpliter(metaclass=ABCMeta):
    def __init__(self, data, train_ratio, valid_ratio, test_ratio, source_type, seed = config.DATA_RANDOM_SEED):
        if not isinstance(data, (dict, list)):
            raise TypeError("data_source has a wrong type, required: list, actually: {}".format(
                type(data).__name__))

        check_split_ratio(train_ratio, valid_ratio, test_ratio)
        self.source_type = source_type
        if not data:
            self.alldata = []
        elif self.source_type in DataSourceType.CV_TYPE_LIST:
            # 图片数据源
            self.alldata = data
        else:
            self.alldata = data.pop("result")
            self.column_meta = data.pop("meta")
        self.data_num = len(self.alldata)
        self.train_num, self.valid_num, self.test_num = get_set_count(
            self.data_num, train_ratio, valid_ratio, test_ratio
        )
        self.train_data = None
        self.valid_data = None
        self.test_data = None

        self.seed = seed  # 使用固定 seed 保证同一个数据集多次读取划分一致

    def train_reader(self) -> Iterable:
        if self.train_num == 0:
            return []
        return self._train_reader()

    @abstractmethod
    def _train_reader(self) -> Iterable:
        pass

    def validation_reader(self) -> Iterable:
        if self.valid_num == 0:
            return []
        return self._validation_reader()

    @abstractmethod
    def _validation_reader(self) -> Iterable:
        pass

    def test_reader(self) -> Iterable:
        if self.test_num == 0:
            return []
        return self._test_reader()

    @abstractmethod
    def _test_reader(self) -> Iterable:
        pass


class NoSpliter(DataSetSpliter):
    """
    不进行数据切分
    """
    def __init__(self, data, train_ratio, valid_ratio, test_ratio, source_type):
        super().__init__(data, train_ratio,
                         valid_ratio, test_ratio, source_type)

    def _train_reader(self):
        return {"meta": self.column_meta, "result": self.alldata}

    def _validation_reader(self):
        return None

    def _test_reader(self):
        return None


class SequentialSpliter(DataSetSpliter):
    """顺序读取数据
    数据按照训练集比例分割，前面为训练集，后面为验证集
    """

    def __init__(self, data, train_ratio, valid_ratio, test_ratio, source_type):
        super().__init__(data, train_ratio,
                         valid_ratio, test_ratio, source_type)

    def _train_reader(self):
        if self.train_data is None:
            self._set_data()
        return self.train_data

    def _validation_reader(self):
        if self.valid_data is None:
            self._set_data()
        return self.valid_data

    def _test_reader(self):
        if self.test_data is None:
            self._set_data()
        return self.test_data

    def _set_data(self):
        # 获取所有数据
        train_data = self.alldata[:self.train_num]
        if self.test_num == 0:
            valid_data = self.alldata[self.train_num:]
            test_data = []
        else:
            valid_data = self.alldata[self.train_num:
                                      self.train_num + self.valid_num]
            test_data = self.alldata[self.train_num + self.valid_num:]

        if self.source_type in DataSourceType.CV_TYPE_LIST:
            self.train_data = train_data
            self.valid_data = valid_data
            self.test_data = test_data
        else:
            self.train_data = {"meta": self.column_meta, "result": train_data}
            self.valid_data = {"meta": self.column_meta, "result": valid_data}
            self.test_data = {"meta": self.column_meta, "result": test_data}


# ShuffleSpliter 乱序读取
class ShuffleSpliter(DataSetSpliter):
    """乱序读取

    注意:
    此方法需要读取全部数据，会给内存带来压力
    """

    def __init__(self, data, train_ratio, valid_ratio, test_ratio, source_type):
        super().__init__(data, train_ratio,
                         valid_ratio, test_ratio, source_type)

    def _train_reader(self):
        if self.train_data is None:
            self._set_data()
        return self.train_data

    def _validation_reader(self):
        if self.valid_data is None:
            self._set_data()
        return self.valid_data

    def _test_reader(self):
        if self.test_data is None:
            self._set_data()
        return self.test_data

    def _set_data(self):
        # 获取所有数据
        r = Random(self.seed)
        r.shuffle(self.alldata)
        train_data = self.alldata[:self.train_num]
        if self.test_num == 0:
            valid_data = self.alldata[self.train_num:]
            test_data = []
        else:
            valid_data = self.alldata[self.train_num:
                                      self.train_num + self.valid_num]
            test_data = self.alldata[self.train_num + self.valid_num:]

        if self.source_type in DataSourceType.CV_TYPE_LIST:
            self.train_data = train_data
            self.valid_data = valid_data
            self.test_data = test_data
        else:
            self.train_data = {"meta": self.column_meta, "result": train_data}
            self.valid_data = {"meta": self.column_meta, "result": valid_data}
            self.test_data = {"meta": self.column_meta, "result": test_data}


class TargetRandomSpliter(DataSetSpliter):
    """根据目标列乱序读取
    按照目标列中类别的比例，进行训练集和验证集的划分，保证训练集和验证集中类别比例与整体数据比例相同

    使用此方法的前提：
     - 有目标列
     - 是分类

    几种可能存在的边界：
     - 分类太多: 1w数据分出来5k类别, 算法框架在tag_count/total > 0.5的时候会warn
     - 类别唯一: 只有一个tag
     - 某类别唯一: 某个tag只有一
     - 无目标列下标: 需要配置targetCol
     - 训练集或验证集比例为1

    注意:
    此方法需要读取全部数据，会给内存带来压力
    """

    def __init__(self, data, train_ratio, valid_ratio, test_ratio, target_col, source_type):
        super().__init__(data, train_ratio,
                         valid_ratio, test_ratio, source_type)

        self.train_ratio = train_ratio
        self.valid_ratio = valid_ratio
        self.test_ratio = test_ratio
        self.target_col = self._target_col(target_col)

    def _target_col(self, target_col):
        if self.source_type in DataSourceType.CV_TYPE_LIST:
            return 'labels'
        return target_col

    def _train_reader(self):
        if self.train_data is None:
            self._set_data()
        return self.train_data

    def _validation_reader(self):
        if self.valid_data is None:
            self._set_data()
        return self.valid_data

    def _test_reader(self):
        if self.test_data is None:
            self._set_data()
        return self.test_data

    def _set_data(self):
        from intelliw.utils.spark_process import Engine
        engine = Engine()

        df = engine.DataFrame(self.alldata)
        if self.source_type in DataSourceType.CV_TYPE_LIST:
            target_index = self.target_col
        else:
            target_index = df.columns[self.target_col]
            # 边界处理
            if df.shape[1] < self.target_col + 1:
                raise DatasetException(
                    f"数据集不存在目标列, 数据集列数{df.shape[1]}, 目标列下标{self.target_col}")

        tags = df[target_index].unique().tolist()
        # 边界处理
        for tag in tags:
            if tag is None or engine.isna(tag):
                tags.remove(tag)
                # break

        tag_count = len(tags)
        if (tag_count < 2) or (tag_count / self.data_num > 1 / 3):
            raise DatasetException(
                f"目标列类别数量唯一, 或者类别数量超过总数据的30%, 类别数量: {tag_count}")

        train, valid, test = engine.DataFrame(), engine.DataFrame(), engine.DataFrame()
        for tag in tags:
            tag_data = df[df[target_index] == tag]

            line_count = tag_data.shape[0]
            train_num, valid_num, test_num = get_set_count(
                line_count, self.train_ratio, self.valid_ratio, self.test_ratio
            )

            # 边界处理
            if (test_num == 0 and line_count < 2) or (test_num != 0 and line_count < 3):
                raise DatasetException(
                    f"目标列类别: {tag} 存在过少的数据({line_count}条), 无法均匀分配到数据集中")

            # train
            tag_data = tag_data.sample(frac=float(1), random_state=self.seed)
            train_data = tag_data[:train_num]
            if test_num == 0:
                valid_data = tag_data[train_num:]
            else:
                valid_data = tag_data[train_num: train_num + valid_num]
                test_data = tag_data[train_num + valid_num:]
                test = engine.concat([test, test_data], ignore_index=True)
            valid = engine.concat([valid, valid_data], ignore_index=True)
            train = engine.concat([train, train_data], ignore_index=True)

        if self.source_type in DataSourceType.CV_TYPE_LIST:
            self.train_data = train.to_dict(orient='records')
            self.valid_data = valid.to_dict(orient='records')
            self.test_data = test.to_dict(orient='records')
        else:
            self.train_data = {"meta": self.column_meta,
                               "result": train.values.tolist()}
            self.valid_data = {"meta": self.column_meta,
                               "result": valid.values.tolist()}
            self.test_data = {"meta": self.column_meta,
                              "result": test.values.tolist()}



#!/usr/bin/env python
# coding: utf-8
import asyncio
import inspect
import json
import threading
import traceback
import importlib
import os
import time
import importlib.util
import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent import futures
from inspect import signature
from typing import Tuple, Union
from intelliw.config.cfg_parser import load_config
from intelliw.datasets.datasets import DataSets, MultipleDataSets
from intelliw.utils import exception, yms_downloader, yms_downloader_v2
from intelliw.utils.exception import ModelLoadException, LimitConcurrencyError
from intelliw.utils import prepare_algorithm_parameters, prepare_model_parameters, import_code, get_first_element
from intelliw.utils.check_algorithm_py import is_valid_algorithm
import intelliw.utils.message as message
from intelliw.core.recorder import Recorder
from intelliw.core.linkserver import linkserver
from intelliw.utils import iprofile
from intelliw.utils.context._context import CurrentReq, header_ctx
from intelliw.utils.logger import _get_framework_logger, _get_algorithm_logger
from intelliw.config import config
from intelliw.utils.global_val import gl
from intelliw import functions
from intelliw.core.al_decorator import make_decorators_server
from intelliw.utils import hijack_function

logger = _get_framework_logger()


class Const:
    # 文件
    algorithmInformation = "AlgorithmInformation"
    algorithm = "algorithm"
    algorithm_py = "algorithm.py"
    algorithm_yaml = "algorithm.yaml"
    model_yaml = "model.yaml"

    # 处理过程
    pre_predict = "pre-predict"
    post_predict = "post-predict"
    pre_train = "pre-train"
    post_train = "post-train"

    # 其他
    new_target_name = "new_target_data_unrepeatable_name"
    metadata = "metadata"
    model_type = "modelType"
    predict_type = "predictType"
    target = "target"
    target_name = "targetName"
    target_col = "targetCol"
    target_code = "targetCode"

    link_server = "linkserver"

    parameters = 'parameters'
    parametersUp = 'PARAMETERS'
    system = 'system'
    delivery = 'delivery'


def get_model_yaml():
    return Const.model_yaml


class Pipeline:
    def __init__(self, reporter: Tuple[str, Recorder] = None):
        self.alldata_transforms = None  # 自定义针对总体数据结构处理的  特征工程
        self.splitdata_transforms = None  # 自定义针对单条数据数据结构处理的  特征工程
        self.feature_transforms = None  # 内置针对总体数据结构处理的  特征工程
        self.recorder = reporter if isinstance(reporter, Recorder) else Recorder(
            reporter)
        self.instance = None
        self.custom_router = []
        self.alg_describe = ''
        self.background_tasks = set()
        self.async_executor = ThreadPoolExecutor(999)
        self.max_wait_task = 999
        self.thread_data_pool = {}

    @staticmethod
    def _check_certified_image():
        # 检查是不是规范化镜像（onnx）
        if any([importlib.util.find_spec("torch"),
                importlib.util.find_spec("paddle"),
                importlib.util.find_spec("tensorflow"),
                not importlib.util.find_spec("onnxruntime")]):
            raise exception.UnverifiedRuntimeException()

    @staticmethod
    def _check_algorithm_parameters_format(alg):
        try:
            alg_ps = alg['parameters'] if 'parameters' in alg else None
            if alg_ps is not None:
                for _, v in enumerate(alg_ps):
                    _ = v['key'] if 'key' in v else None
                    option = v['option'] if 'option' in v else None
                    if option is None:
                        return False
            return True
        except:
            return False

    def _prepare(self, path):
        sys.path.append(path)
        py_file = os.path.join(path, Const.algorithm_py)
        spec = importlib.util.spec_from_file_location(Const.algorithm, py_file)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        return self.module

    @staticmethod
    def load_config(yaml_file):
        cfg = load_config(yaml_file)
        return cfg

    @staticmethod
    def debug_mode():
        if os.environ.get('INTELLIW_HOLD_DEBUG') == "1":
            logger.warn("Intelliw will hold for one hour for debugging, please clear the container after use")
            time.sleep(3600)
            return True
        return False

    @staticmethod
    def initialize_env(envs):
        if envs and "parameters" in envs:
            ps: dict = envs.get("parameters", {})
            for k, v in ps.items():
                os.environ[str(k)] = str(v)

    def _initialize_algorithm(self, algorithm_yaml_file):
        cfg = self.load_config(algorithm_yaml_file)
        gl.alg_yaml_file = algorithm_yaml_file

        # 路由配置获取
        self.custom_router = cfg[Const.algorithmInformation].get(
            'router', [])

        # 算法配置获取
        alg_ps = cfg[Const.algorithmInformation][Const.algorithm].get(
            'parameters', None)
        return prepare_algorithm_parameters(alg_ps)

    def _initialize_model(self, model_yaml_file, parameters, option):
        model_cfg = self.load_config(model_yaml_file)
        self.model_cfg = model_cfg['Model']

        # 手动写入基准模型路径, 默认是 ./base_model
        if config.BASE_MODEL_MODE and self.model_cfg.get('base_model_location', None) is None:
            self.model_cfg['base_model_location'] = config.BASE_MODEL_LOCATION

        # 算法描述
        self.alg_describe = model_cfg['Model'].get('desc', '')

        # 算法元数据
        metadata = self.model_cfg.get(Const.metadata, {})

        # 算法类型 string（分类回归时间序列, classify|regression）
        self.predict_type = metadata.get(Const.predict_type)
        # 算法类型 int（分类回归时间序列。。。）
        self.model_type = metadata.get(Const.model_type)
        if self.model_type is not None:
            gl.model_type = self.model_type
        gl.model_yaml_file = model_yaml_file

        model_ps = self.model_cfg.get(Const.parameters)

        # model and algorithm's parameters are all in this var 'parameters'
        parameters = prepare_model_parameters(
            model_ps, parameters, option)

        # 从环境变量中加载PARAMETERS
        env_model_ps = os.environ.get(Const.parametersUp)
        if env_model_ps:
            env_model_ps = json.loads(env_model_ps)
            parameters = prepare_model_parameters(
                env_model_ps, parameters, option)

        # 获取系统级环境变量
        sys_envs = self.model_cfg.get(Const.system)
        self.initialize_env(sys_envs)

        if self.debug_mode():
            exit(0)

        # 特征处理可能会修改target_col，所以需要先保存起来, 有target_name时一定会有target_col
        target_metadata_cfg = metadata.get(Const.target, [])
        target_metadata = None
        if len(target_metadata_cfg) > 0:
            target_metadata = [{"target_col": int(m[Const.target_col]), "target_name": m.get(
                Const.target_code, m[Const.target_name])} for m in target_metadata_cfg]
        else:
            # TODO 暂时兼容 后面删了
            target_col = parameters.get(Const.target_col)
            target_name = parameters.get(Const.target_name)
            if target_col is not None or target_name is not None:
                target_col = int(
                    target_col) if target_col is not None else None
                target_metadata = [
                    {"target_col": target_col, "target_name": target_name}]

        gl.target_metadata = target_metadata

        # link server
        linkserver.config = self.model_cfg.get(Const.link_server, None)

        # self config
        gl.self_config = self.model_cfg.get(Const.delivery, None)

        # *aidataset* delivery 内容写入 parameters
        if gl.self_config:
            for i in gl.self_config:
                # 如果key已经存在，则打印旧值和新值, 继续覆盖
                if parameters.get(i['key']) is not None:
                    logger.info( f"self_config key: {i['key']} old value: {parameters[i['key']]} new value: {i['val']}")
                parameters[i['key']] = i['val']

        return parameters, option

    def initialize_config(self, alg_file, model_file=None):
        # algorithm.yaml
        parameters, option = self._initialize_algorithm(alg_file)

        # model.yaml
        if model_file is not None:
            parameters, option = self._initialize_model(
                model_file, parameters, option)
        return parameters, option

    def _initialize(self, path, alg_file, model_file=None):
        try:
            # 初始化配置
            parameters, option = self.initialize_config(alg_file, model_file)

            # 校验
            if self._prepare(path) is None:
                self.raise_exception(
                    message.err_import_alg_invalid_algorithpy_format)
            is_valid, msg = is_valid_algorithm(self.module)
            if not is_valid:
                self.raise_exception(msg)

            # parameters
            # 日志
            parameters["framework_log"] = _get_algorithm_logger()
            if self.predict_type:
                parameters[Const.predict_type] = self.predict_type

            instance = self.module.Algorithm(parameters)
            if instance is None:
                self.raise_exception(message.err_import_alg_invalid_path)

            # 检测自定义路由方法是否存在
            for router in self.custom_router:
                func = router.get('func')
                if not hasattr(instance, func):
                    self.raise_exception(message.err_missing_router_func)

            # add decorator to functions
            if config.is_server_mode():
                make_decorators_server(instance, self.recorder)

            # FRAMEWORK_MODE
            instance.is_train_mode = config.FRAMEWORK_MODE == config.FrameworkMode.Train
            instance.is_infer_mode = config.FRAMEWORK_MODE == config.FrameworkMode.Infer

            return instance
        except Exception as e:
            self.raise_exception(message.err_import_alg_invalid_path, e)

    def _importalg_phase(self, path):
        hijack_function.HijackRequests()
        # 镜像校验
        if config.CHECK_CERTIFIED_IMAGE:
            hijack_function.HijackModelLoad()

        if not os.path.isfile(os.path.join(path, Const.algorithm_py)):
            self.raise_exception(message.err_import_alg_missing_algorithpy)
        if not os.path.isfile(os.path.join(path, Const.algorithm_yaml)):
            self.raise_exception(message.err_import_alg_missing_algorithyaml)
        cfg = self.load_config(os.path.join(path, Const.algorithm_yaml))
        self.algorithm_cfg = cfg  # SELF-DEFINE DICT
        if not self._check_algorithm_parameters_format(cfg):
            self.raise_exception(
                message.err_import_alg_invalid_algorithyaml_format)

    def _load_model(self, path, model_path):
        model_file = model_path if os.path.isabs(model_path) else \
            os.path.join(path, model_path.lstrip("./"))
        logger.info(f"model load path：{model_file}")
        self.instance.load(model_file)

    def _load_base_model(self, path, base_model_path):
        base_model_path = base_model_path if os.path.isabs(base_model_path) else \
            os.path.join(path, base_model_path.lstrip("./"))
        logger.info(f"base model load path：{base_model_path}")
        self.instance.load(base_model_path)

    def importalg(self, path, need_initialize=True):
        """
        导入算法
        :param path: 算法包所在路径
        :return: None
        :raises PipelineException: 框架导入算法失败原因
        """
        try:
            self._importalg_phase(path)
            if need_initialize:
                self._initialize(os.path.join(path, Const.algorithm_yaml))
        except Exception as e:
            self.raise_exception(message.err_import_algorithm, e)
        self.recorder.report(message.ok_import_algorithm)

    def importmodel(self, path, need_initialize=True, load_model=True, load_base_model=False):
        """
        导入模型
        :param path: 算法包所在路径
        :param load_model: 是否要导入模型，即是否调用 Algorithm 的 load 方法
        :return: None
        :raises PipelineException: 框架导入模型失败原因
        """
        model_yaml_path = os.path.join(path, Const.model_yaml)
        algorithm_yaml_path = os.path.join(path, Const.algorithm_yaml)
        # check model yaml file
        if not os.path.isfile(model_yaml_path):
            logger.error(
                "未找到 model.yaml, path: {}, file_name: {}".format(path, Const.model_yaml))
            self.raise_exception(message.err_import_model_missing_modelyaml)

        try:
            # check algorithm package
            self._importalg_phase(path)
            if need_initialize:
                self.instance = self._initialize(
                    path, algorithm_yaml_path, model_yaml_path)

                # get transforms
                trans_cfg = self.model_cfg.get('transforms', None)
                transforms_result = self.prepare_transforms(trans_cfg)
                self.splitdata_transforms = transforms_result[0]
                self.alldata_transforms = transforms_result[1]
                self.feature_transforms = transforms_result[2]
                gl.set_dict({
                    'feature_process': self.feature_transforms,
                    'timeseries_process': transforms_result[3]
                })
                
                # 判断是否使用 model manager 管理模型
                # 条件1: algorithm 实例中包含 model_manager_mode 属性，且值为 True,
                # 条件2: algorithm 实例中包含 model_manager 属性，且值为 ModelManager 实例
                # 条件3: algorithm 实例中包含 load_with_manager 方法，且为方法
                # 条件4: algorithm 实例中包含 is_infer_mode 属性，且值为 True
                load_model_manager = False
                if hasattr(self.instance, 'is_infer_mode') and self.instance.is_infer_mode is True and  \
                 hasattr(self.instance, 'model_manager_mode') and self.instance.model_manager_mode is True and \
                 hasattr(self.instance, 'model_manager') and  hasattr(self.instance, 'load_with_manager') and \
                inspect.ismethod(self.instance.load_with_manager):
                    load_model_manager = True

                # load model manager
                if load_model_manager:
                    try:
                        model_path: str = self.model_cfg['location']
                        self._initial_model_manager(path, model_path)
                    except Exception as e:
                        raise ModelLoadException(e)
                # load model 优先加载 model
                elif load_model:
                    try:
                        model_path: str = self.model_cfg['location']
                        self._load_model(path, model_path)
                    except Exception as e:
                        raise ModelLoadException(e)
                elif load_base_model:
                    try:
                        base_model_path: str = self.model_cfg['base_model_location']
                        self._load_base_model(path, base_model_path)
                    except Exception as e:
                        raise ModelLoadException(e)

                del self.model_cfg
        except Exception as e:
            self.raise_exception(message.err_import_model, e)

        # report
        self.recorder.report(message.ok_import_model)

    def _vaild_transforms_cfg(self, stage, is_split):
        transforms = self.splitdata_transforms if is_split else self.alldata_transforms
        if transforms is not None:
            transforms = transforms.get(stage)
            if transforms:
                return transforms
        return None

    def _transform_helper(self, stage, data, is_split):
        transforms = self._vaild_transforms_cfg(stage, is_split)
        if transforms is not None:
            return self.transform_call_chains(transforms, data)
        return data

    def _transform_function(self, stage, is_split):
        if self._vaild_transforms_cfg(stage, is_split) is None:
            return None

        def _transform(data):
            return self._transform_helper(stage, data, is_split)

        return _transform

    def _valid_feature_transforms(self, stage):
        if self.feature_transforms is not None:
            params = self.feature_transforms.get(stage)
            if params:
                return params
        return None

    def _feature_helper(self, stage, data):
        params = self._valid_feature_transforms(stage)
        if params is not None:
            from intelliw.functions.feature_process import feature_process
            return feature_process(data, params, stage)
        return data

    def _feature_process(self, stage):
        if self._valid_feature_transforms(stage) is None:
            return None

        def _process(data):
            return self._feature_helper(stage, data)

        return _process

    def _get_train_kwargs(self, reader, column_meta):
        if reader is None:
            return {}

        # config
        feature_process_cfg = gl.pop("feature_process")
        if feature_process_cfg is not None:
            feature_process_cfg = {
                Const.pre_train: feature_process_cfg.get(Const.pre_train),
                Const.post_train: feature_process_cfg.get(Const.post_train)
            }

        final_target_col = gl.pop("final_target_col")
        target_cols = gl.pop("target_cols")
        target_metadata = gl.pop("target_metadata")
        if target_cols is None and target_metadata is not None and len(target_metadata) > 0:
            target_cols = [
                {
                    "col": m["target_col"],
                    "column_name": m["target_name"]
                } for m in target_metadata
            ]
            final_target_col = target_cols[-1]["col"]

        kwargs = {
            "final_target_col": final_target_col,
            "target_cols": target_cols,
            "category_cols": gl.pop("category_cols"),
            "column_meta": gl.pop("column_meta") if gl.column_meta is not None else [i["code"] for i in
                                                                                     column_meta],
            "column_relation_df": gl.pop("column_relation_df"),
            "feature_process_cfg": feature_process_cfg,
        }

        # self config
        if gl.self_config:
            kwargs = {**kwargs, **{i['key']: i['val'] for i in gl.self_config}}

        return kwargs

    def train(self, datasets: Union[DataSets, MultipleDataSets]):
        """
        训练入口
        """
        if datasets is None:
            self.raise_exception(message.err_empty_train_data)

        # 特征值处理
        try:
            if config.SOURCE_TYPE == 0:
                train_set, valid_set, test_set = [], [], []
            else:
                train_set, valid_set, test_set = datasets.data_pipeline(
                    self._transform_function(Const.pre_train, True),
                    self._transform_function(Const.pre_train, False),
                    self._feature_process(Const.pre_train)
                )
        except Exception as e:
            self.raise_exception(message.err_function_process, e)

        # 开始训练
        try:
            logger.info("data process success, start train mode")

            # 数据集
            self.instance.train_set = train_set
            self.instance.valid_set = valid_set
            self.instance.test_set = test_set

            # 训练参数
            train_func = getattr(self.instance, "train")
            varnames = signature(train_func).parameters
            if "kwargs" not in varnames:
                # train(train, valid)
                train_func(
                    self.instance.train_set,
                    self.instance.valid_set
                )
            else:
                kwargs = self._get_train_kwargs(
                    train_set,
                    datasets.column_meta
                )
                if len(varnames) == 1:
                    # train(**kwargs)
                    train_func(**kwargs)
                elif len(varnames) == 3:
                    # train(train_set, valid_set, **kwargs)
                    train_func(
                        self.instance.train_set,
                        self.instance.valid_set,
                        **kwargs
                    )
                else:
                    self.raise_exception(message.err_missing_train_parameter)
            self.recorder.report(message.ok_train_finish)
        except exception.DatasetException as de:
            self.raise_exception(
                message.err_dataset, de, is_framework_error=False
            )
        except Exception as e:
            self.raise_exception(
                message.err_train, e, is_framework_error=False
            )

    def train_streaming(self, datasets: Union[DataSets, MultipleDataSets]):
        """
        流式训练入口 - 支持大数据集按列、分片处理
        透传数据集 datasets
        """

        logger.info("start streaming train mode")
        try:
            # 训练参数
            if not hasattr(self.instance, "train_streaming"):
                self.raise_exception(message.err_streaming_not_supported)
                
            train_streaming_func = getattr(self.instance, "train_streaming")
            varnames = signature(train_streaming_func).parameters
            
            # 检查算法是否支持流式训练
            supports_streaming = "datasource" in varnames
            if not supports_streaming:
                logger.error("算法未实现流式训练接口, train_streaming(datasource)")
                self.raise_exception(message.err_streaming_not_supported)
            
            if len(varnames) == 1:
                # train_streaming(datasource)
                train_streaming_func(datasets.datasource)
            else:
                self.raise_exception(message.err_missing_train_parameter)
            # 流式训练完成信号
            self.recorder.report(message.ok_train_finish)
            
        except exception.DatasetException as de:
            self.raise_exception(
                message.err_dataset, de, is_framework_error=False
            )
        except Exception as e:
            self.raise_exception(
                message.err_train, e, is_framework_error=False
            )
    async def _infer_process(self, data, request, func='infer', need_feature=False):
        work_queue_size = self.async_executor._work_queue.qsize()
        if func == 'intelliw-worker':
            return {
                "num_threads": len(self.async_executor._threads),
                "num_process": config.INFER_MULTI_PROCESS_COUNT,
                "cur_process": os.getpid(),
                "working_task": work_queue_size,
                "max_wait_task/process": self.max_wait_task,
                "total_wait_task": self.max_wait_task * config.INFER_MULTI_PROCESS_COUNT
            }

        if work_queue_size >= self.max_wait_task:
            error_msg = (f"Exceeded concurrency limit, "
                         f"Max wait task limit: {self.max_wait_task * config.INFER_MULTI_PROCESS_COUNT}")
            logger.warning(error_msg)
            raise LimitConcurrencyError(error_msg)

        loop = asyncio.get_event_loop()

        if need_feature:  # 前处理
            data = await loop.run_in_executor(
                self.async_executor, self._feature_helper, Const.pre_predict,
                self._transform_helper(Const.pre_predict, data, False)
            )

        # profile
        check_profile = request.query.get('intelliw_profile')
        check_mprofile = request.query.get('intelliw_mprofile')

        # 推理
        infer_func = getattr(self.instance, func)
        arg = [data] if infer_func.__code__.co_argcount > 1 else []  # 获取函数参数数量

        if not inspect.iscoroutinefunction(infer_func):

            cur_header_ctx = header_ctx.get()

            def infer_process(*a):
                self.instance.request = CurrentReq(request)
                header_ctx.set(cur_header_ctx)

                # 数据池
                if self.thread_data_pool:
                    t = threading.currentThread().ident
                    self.instance.request.datapool = self.thread_data_pool.get(t)

                if check_profile:  # 性能检测
                    return iprofile.performance_profile(infer_func, *a)
                elif check_mprofile:  # 内存检测
                    try:
                        from intelliw.utils import memory_profiler
                    except ImportError as e:
                        return {"result": e}
                    memory_usage, res = memory_profiler.memory_usage(
                        (infer_func, a), retval=True,
                        include_children=True,
                        interval=.2, max_iterations=1)
                    return {"result": res, "profile": memory_usage}
                else:
                    return infer_func(*a)

            # 非异步算法异步执行
            result = await loop.run_in_executor(
                self.async_executor, infer_process, *arg
            )
        else:
            self.instance.request = CurrentReq(request)

            # 异步算法默认已经做了从头到尾的异步处理
            # 出现阻塞问题，是算法对async await理解不够透彻
            if check_profile:
                result = await iprofile.async_performance_profile(infer_func, *arg)
            else:
                result = await infer_func(*arg)

        if need_feature:  # 后处理
            result = await loop.run_in_executor(
                self.async_executor, self._transform_helper, Const.post_predict, result, False
            )

        return result

    async def infer(self, data, request, func='infer', need_feature=True):
        """
        推理入口
        """
        try:
            start = time.time()

            result = await self._infer_process(
                data, request, func, need_feature
            )

            rst = request.raw.state.req_start_time
            ist = request.raw.state.ingress_start_time
            ingress_duration = rst - float(ist) / 1000 if ist else 0
            req_duration = start - rst if rst else None

            logger.info(
                f'{request.method} {request.url.path if request.url else "TestCase"}   function: {func}   '
                f'time: \033[33m connect: {ingress_duration:.4f}s, request: {req_duration:.4f}s, '
                f'response: {time.time() - start:.4f}s\033[0m '
            )
            return result, None
        except LimitConcurrencyError as e:
            msg = f"inference error: {e}"
            return f"inference error: {e}", LimitConcurrencyError()
        except Exception as e:
            stack_info = traceback.format_exc()
            logger.error("推理错误 %s, stack:\n%s", e, stack_info)
            return None, f"inference error: {e}"

    def transform_call_wrapper(self, src, func, params):
        """
        数据处理调用
        """
        if type(func) is type:
            return func().__call__(src, params)
        return func(src, params)

    def transform_call_chains(self, funcs, src):
        """
        数据处理函数链
        data, config -> func1(data, config) -> data, config -> func2(data, config) -> ...
        """
        name = ""
        try:
            for _, item in enumerate(funcs):
                (name, func, params) = item
                if id(func) == id(functions.select_columns) and \
                        self._not_select_columns():
                    continue
                src = self.transform_call_wrapper(src, func, params)
            return src
        except Exception as e:
            msg = message.CommonResponse(
                500, "transforms", f"特征工程函数 {name} 出错")
            self.raise_exception(msg, e)

    def _not_select_columns(self):
        """
        如果有特征工程前处理，就需要忽略select_columns
        """
        if gl.feature_process and gl.feature_process.get(Const.pre_train):
            return True
        return False

    def _get_func_by_name(self, name):
        """
        根据 name 加载函数

        加载顺序:
        1. 尝试在算法项目中加载
        2. 尝试在内置的特征处理中查找
        3. 尝试在当前 module 中加载

        :param name: 待加载的函数名称
        :return: 函数
        :raise: RuntimeError 加载失败原因
        """
        # 尝试在算法项目中加载
        mod = self.module
        if hasattr(mod, name):
            return getattr(mod, name)

        # 尝试在内置的特征处理中查找
        if hasattr(functions, name):
            return getattr(functions, name)

        # 尝试在当前 module 中加载
        mod = sys.modules[__name__]
        if hasattr(mod, name):
            return getattr(mod, name)

        raise RuntimeError(f'函数无效: [{name}]')

    @staticmethod
    def _prepare_timeseries_transforms(timeseries_features, typ, params):
        # normal
        normal = json.loads(params.get(
            functions.Const.timeseries_param, 'null')
        )
        if normal and normal.get("holiday", "none") == "none":
            normal["holiday"] = None
        timeseries_features[typ]["normal"] = normal

        # group
        group_params = json.loads(params.get(
            functions.Const.timeseries_group_param, '[]')
        )
        for idx, i in enumerate(group_params):
            k = f"industry_{idx}" if idx > 0 else "group"
            timeseries_features[typ][k] = i
        return timeseries_features

    def prepare_transforms(self, cfgs):
        """
        解析配置文件中的特征工程
        """

        def init_dict(item):
            return {} if item is None else item

        def init_list(item):
            return [] if item is None else item

        # 没用到就不初始化，减少后面的判断
        split_transforms, alldate_transforms = None, None
        data_features, timeseries_features = None, None
        if not cfgs:
            return split_transforms, alldate_transforms, data_features, timeseries_features

        for cfg in cfgs:
            typ, functions_cfg = cfg.get('type'), cfg.get('functions')
            if typ is not None and functions_cfg is not None:
                for item in functions_cfg:
                    key = item.get('key')
                    params, _ = prepare_algorithm_parameters(
                        item.get('parameters')
                    )
                    if key == functions.Const.feature_process:
                        # 特征工程
                        # 如果是离职预测，所有的特征auto配置必须为0，只做列数据筛选
                        data_features = init_dict(data_features)
                        data_features[typ] = json.loads(params.get(
                            functions.Const.feature_process_param, 'null')
                        )
                    elif key == functions.Const.timeseries_process:
                        # 时间序列
                        timeseries_features = init_dict(
                            timeseries_features)
                        timeseries_features[typ] = dict()
                        timeseries_features = self._prepare_timeseries_transforms(
                            timeseries_features, typ, params
                        )
                    else:
                        # 常规数据处理
                        body = item.get('body')
                        func = self._get_func_by_name(
                            key) if body is None else get_first_element(import_code(body, None))
                        if func is not None:
                            can_split = key in functions.Const.split_function_list or item.get(
                                'canSplit')

                            _t = {}
                            if can_split:
                                split_transforms = init_dict(
                                    split_transforms)
                                _t = split_transforms
                            else:
                                alldate_transforms = init_dict(
                                    alldate_transforms)
                                _t = alldate_transforms

                            _t[typ] = init_list(_t.get(typ))
                            _t[typ].append((key, func, params))
                            logger.info(
                                "特征函数 %s, 是否可以进行分片处理 %s", key, can_split)
                        else:
                            logger.info("找不到函数 %s", key)
        return split_transforms, alldate_transforms, data_features, timeseries_features

    def validate_transforms(self, cfgs, data):
        """
        验证特征工程函数， 用于用户校验函数正确性和结果，服务于validateService
        """
        transforms = []
        if not cfgs:
            return self.transform_call_chains(transforms, data)

        for cfg in cfgs:
            transforms = self._set_transform(cfg, transforms)

        return self.transform_call_chains(transforms, data)

    def _set_transform(self, cfg, transforms):
        inner_functions = cfg.get('functions', None)
        if not inner_functions:
            return transforms

        for item in inner_functions:
            desc = item['desc'] if 'desc' in item else None
            key = item['key'] if 'key' in item else None
            body = item['body'] if 'body' in item else None
            parameters = item['parameters'] if 'parameters' in item else None
            if body is not None:
                n_mod = import_code(body, None)
                func = get_first_element(n_mod)
            else:
                func = self._get_func_by_name(key)
            params, _ = prepare_algorithm_parameters(parameters)
            if func is not None:
                logger.info("特征函数 %s", key)
                transforms.append((key, func, params))
            else:
                logger.info("找不到函数 %s", key)
        return transforms

    def raise_exception(self, msg: message.CommonResponse,
                        error: Exception = None, is_framework_error=True):
        raise_logger = logger if is_framework_error else \
            _get_algorithm_logger()

        # 用于传递错误给prepare_env.sh
        err_stack = "\n"
        if hasattr(error, "ignore_stack") or isinstance(error, ImportError):
            logger.error(traceback.format_exc())
            err_stack += f"错误描述: {error}"  # 数据集错误不需要调用栈，知道什么问题就行
        elif error:
            err_stack += f"错误描述: {error} \n错误详情: {traceback.format_exc()}"
        else:
            err_stack = ""
        err_content = f"\n错误内容: {msg.msg}{err_stack}"
        os.environ["ERR_MASSAGE"] = err_content

        if error is not None:
            msg.msg = err_content
        raise_logger.error(msg.msg)

        if hasattr(error, "btype"):
            msg.other['businessType'] = error.btype()

        # 上报
        self.recorder.report(msg)
        sys.exit()

    def _initial_model_manager(self, path, model_path):
        # 支持ModelManager参数透传
        model_file = model_path if os.path.isabs(model_path) else \
            os.path.join(path, model_path.lstrip("./"))
        try:
            from intelliw.utils.modelmanager.model_manager import get_model_manager
            model_manager = get_model_manager(model_file, self.instance.load_with_manager)
            self.instance.model_manager = model_manager
            """加载模型，支持ModelManager动态获取和延迟切换"""
            logger.info(f"model load path：{model_file}")
        except Exception as e:
            self.instance.model_manager = None
            logger.error(f"Failed to initialize model manager: {e}")
        
        # 检查是否支持ModelManager
        if hasattr(self.instance, 'model_manager') and self.instance.model_manager is not None:
            self.instance.get_model = self.instance.model_manager.get_active_model
            self.instance.get_model_version = self.instance.model_manager.get_model_version
        


import asyncio
import os
import subprocess
import time
import threading

from intelliw.config import config
from intelliw.utils import stop_thread
from intelliw.utils.global_val import gl
from intelliw.utils.logger import _get_framework_logger
from intelliw.utils.storage_service import FileTransferDevice, StorageService

logger = _get_framework_logger()


def init_request_data(data):
    # 请求参数校验
    if data.get('requestMode') == "monitor":
        monitorTime = data.get('monitorTime')
        # 请求监控必须要传monitorTime
        if monitorTime is None:
            return {"status": 9, "msg": "MonitorTime cannot be None"}
        if not isinstance(monitorTime, int):
            return {"status": 9, "msg": "MonitorTime must be of int type"}
        # monitorTime大小限制
        if monitorTime < 30 or monitorTime > 30 * 60:
            return {"status": 9, "msg": "The size range of monitorTime should be 30-1800"}
    return None


def monitor_query():
    # 默认返回状态为无监控任务
    response_data = {"status": 0}
    # 任务存在
    if gl.monitor:
        if gl.monitor.get("t").is_alive():
            # 监控线程存活且未超时，返回状态监控中
            if time.time() - gl.monitor.get("start") < 30 * 60:
                response_data = {"status": 1}
            # 监控线程存活且超时，杀掉监控线程，返回状态无任务
            else:
                stop_thread(gl.monitor.get("t"))
        # 监控线程不存活且有url，返回状态已完成
        elif gl.monitor.get("url"):
            response_data = {"status": 2, "url": gl.monitor.get("url")}
    return response_data


def monitor_request(data):
    monitorTime = data.get('monitorTime')
    # 无任务，创建新监控
    if not gl.monitor:
        start_new(monitorTime)
    # 有任务
    else:
        # 监控线程存活且未超时，返回状态监控中（默认）
        if gl.monitor.get("t").is_alive():
            # 监控线程存活且未超时，返回状态监控中（默认）
            if time.time() - gl.monitor.get("start") < 30 * 60:
                logger.info("Monitoring thread exists")
            # 监控线程存活且超时，杀掉监控线程，创建新监控，返回状态监控中（默认）
            else:
                stop_thread(gl.monitor.get("t"))
                start_new(monitorTime)
        # 监控线程不存活，创建新监控，返回状态监控中（默认）
        else:
            start_new(monitorTime)


def start_new(monitorTime):
    file_path = f'./result-{int(time.time())}.svg'
    # 异步上传
    thread = threading.Thread(target=upload_file, args=(file_path, monitorTime))
    thread.start()
    # 返回文件路径
    curKey = os.path.join(config.STORAGE_SERVICE_PATH, config.SERVICE_ID, os.path.basename(file_path))
    downloader = StorageService(key=curKey, process_type="download")

    gl.monitor = {"start": time.time(), "url": downloader.service_url, "t": thread}


def upload_file(file_path, monitorTime):
    """
        生成火焰图并且上传文件
    """
    cmdline = ['py-spy', 'record', '-o', file_path, '-s', '-d', str(monitorTime), '-p', config.MAINPID]
    try:
        subprocess.check_output(cmdline)
    except Exception as e:
        logger.info("堆栈火焰图错误： %s", e)
        pass
    FileTransferDevice(
        file_path, "profile_result", os.path.basename(file_path)
    )



import datetime
import typing
import numpy as np

from fastapi.responses import JSONResponse as J, StreamingResponse


class JSONResponse(J):

    def render(self, content: typing.Any):
        if isinstance(content, (np.integer, np.floating, np.bool_)):
            return content.item()
        elif isinstance(content, np.ndarray):
            return content.tolist()
        elif isinstance(content, (datetime.datetime, datetime.timedelta)):
            return content.__str__()
        else:
            return super(J, self).render(content)



import base64
import hashlib
import hmac
import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class YmsEncryptUtil:
    _log = logging.getLogger("YmsEncryptUtil")
    _aes_key_bytes = bytearray(base64.b64decode("g/A/0jw92C1Ki7toxpL6hzAqpYaGd3oUo9OLengo1Ck="))
    _aes_key_iv = bytearray(base64.b64decode("s6jg4fmrG/46NvIx9nb3i7MKUZ0rIebFMMDu6Ou0pdA="))
    _aes_key_aad = bytearray(base64.b64decode("NanjgbGidWdUm1+Kb3g8Fn6/gJ8cTWqeNnJASE2M4oE="))
    _initialized = False

    DEFAULT_FACTORS = ("123", "456", "789")

    CIPHER_TEXT_PREFIX = "YMS("
    CIPHER_TEXT_SUFFIX = ")"
    YMS_BASE64_PREFIX = "BAS("
    YMS_BASE64_SUFFIX = ")"

    @classmethod
    def __init__(cls, factor1: str = None, factor2: str = None, factor3: str = None):
        if not cls._initialized:
            cls._initialized = True
            factors = []
            for f in (factor1, factor2, factor3):
                if not f:
                    factors.extend(cls.DEFAULT_FACTORS)
                    break
                factors.append(cls._decode_encrypt_factor(f))
            else:
                cls._gen_aes_keys(*factors)

    @classmethod
    def _decode_encrypt_factor(cls, factor: str) -> str:
        if factor.startswith(cls.YMS_BASE64_PREFIX) and factor.endswith(cls.YMS_BASE64_SUFFIX):
            decoded = base64.b64decode(factor[len(cls.YMS_BASE64_PREFIX):-len(cls.YMS_BASE64_SUFFIX)])
            return decoded.decode('utf-8')
        return factor

    @classmethod
    def _gen_aes_keys(cls, fact1: str, fact2: str, fact3: str):
        cls._aes_key_bytes = cls._gen_config_file_key(fact1, fact2, fact3)
        try:
            cls._aes_key_iv = hashlib.sha256(fact2.encode()).digest()
            cls._aes_key_aad = hashlib.sha256(fact3.encode()).digest()
        except Exception as e:
            cls._log.warning(f"生成Iv和附加字符失败: {e}")

    @classmethod
    def _gen_config_file_key(cls, str1: str, str2: str, str3: str) -> bytes:
        try:
            data = hashlib.sha256(str3.encode()).digest()
            password = hmac.new(str2.encode(), str1.encode(), hashlib.sha256).digest()

            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=data,
                iterations=10000,
                backend=default_backend()
            )
            return kdf.derive(password)
        except Exception as e:
            cls._log.error(f"生成配置文件密钥失败: {e}")
            return b''

    @classmethod
    def encrypt(cls, s: str) -> str:
        if s == "" or (s.startswith(cls.CIPHER_TEXT_PREFIX) and s.endswith(cls.CIPHER_TEXT_SUFFIX)):
            return s
        return cls._enc_config_file_item(s)

    @classmethod
    def decrypt(cls, s: str) -> str:
        if s == "" or not (s.startswith(cls.CIPHER_TEXT_PREFIX) and s.endswith(cls.CIPHER_TEXT_SUFFIX)):
            return s
        return cls._dec_config_file_item(s)

    @classmethod
    def _enc_config_file_item(cls, plain_text: str) -> str:
        try:
            encryptor = Cipher(
                algorithms.AES(cls._aes_key_bytes),
                modes.GCM(cls._aes_key_iv, min_tag_length=8),
                backend=default_backend()
            ).encryptor()
            encryptor.authenticate_additional_data(cls._aes_key_aad)
            ciphertext = encryptor.update(plain_text.encode()) + encryptor.finalize()
            add_tag = encryptor.tag[:-8]
            return cls.CIPHER_TEXT_PREFIX + (ciphertext + add_tag).hex().upper() + cls.CIPHER_TEXT_SUFFIX
        except Exception as e:
            cls._log.error(f"加密失败: {e}")
            return ""

    @classmethod
    def _dec_config_file_item(cls, cipher_hex: str) -> str:
        cipher_hex = cipher_hex[len(cls.CIPHER_TEXT_PREFIX):-len(cls.CIPHER_TEXT_SUFFIX)]
        try:
            data = bytes.fromhex(cipher_hex)
            tag, ciphertext = data[-8:], data[:-8]

            decryptor = Cipher(
                algorithms.AES(cls._aes_key_bytes),
                modes.GCM(cls._aes_key_iv, tag, min_tag_length=8),
                backend=default_backend()
            ).decryptor()
            decryptor.authenticate_additional_data(cls._aes_key_aad)
            return (decryptor.update(ciphertext) + decryptor.finalize()).decode()
        except Exception as e:
            cls._log.error(f"解密失败: {cipher_hex[:8]}...: {e}")
            return f"{cls.CIPHER_TEXT_PREFIX}{cipher_hex}{cls.CIPHER_TEXT_SUFFIX}"



#!/usr/bin/env python
# coding: utf-8

import inspect
from inspect import signature
from typing import Tuple
import intelliw.utils.message as message


def is_valid_algorithm(mod) -> Tuple[bool, message.CommonResponse]:
    if not inspect.ismodule(mod):
        return False, message.err_missing_algorithm_mod

    if not (hasattr(mod, 'Algorithm') and inspect.isclass(mod.Algorithm)):
        return False, message.err_missing_algorithm_class

    cls = mod.Algorithm
    if not (hasattr(cls, 'train') and inspect.isfunction(cls.train)):
        return False, message.err_missing_train_method
    else:
        sig = signature(cls.train)
        if len(sig.parameters) < 2 or 'self' not in sig.parameters:
            return False, message.err_missing_train_parameter

    if not (hasattr(cls, 'infer') and inspect.isfunction(cls.infer)):
        return False, message.err_missing_infer_method
    else:
        sig = signature(cls.infer)
        if len(sig.parameters) < 1 or 'self' not in sig.parameters:
            return False, message.err_missing_infer_parameter

    if not (hasattr(cls, '__init__') and inspect.isfunction(cls.__init__)):
        return False, message.err_missing_init_method
    else:
        sig = signature(cls.__init__)
        if len(sig.parameters) != 2 or 'self' not in sig.parameters:
            return False, message.err_missing_init_parameter

    if not (hasattr(cls, 'load') and inspect.isfunction(cls.load)):
        return False, message.err_missing_load_method
    else:
        sig = signature(cls.load)
        if len(sig.parameters) != 2 or 'self' not in sig.parameters:
            return False, message.err_missing_load_parameter

    if not (hasattr(cls, 'save') and inspect.isfunction(cls.save)):
        return False, message.err_missing_save_method
    else:
        sig = signature(cls.save)
        if len(sig.parameters) < 1 or 'self' not in sig.parameters:
            return False, message.err_missing_save_parameter

    return True, None



# coding=utf-8
import os
import shutil
from intelliw import __version__
from setuptools import setup, find_packages


def move_base_file(infile, outfile):
    """move intelliw/scaffolding/resources/iw.py to workdir """
    shutil.copy(infile, outfile)


def parse_requirements(path):
    """ load requirements from a pip requirements file. (replacing from pip.req import parse_requirements)"""
    with open(path, 'r', encoding='utf8') as fd:
        return [r.strip() for r in fd.readlines()]


def long_description_load(filename):
    with open(filename, "r", encoding='utf8') as fh:
        long_description = fh.read()
    return long_description


def purge_pycache(path):
    for file_name in os.listdir(path):
        abs_path = os.path.join(path, file_name)
        if file_name in ("__pycache__", "logs"):
            shutil.rmtree(abs_path)
            print(f"delete: {abs_path}")
        elif os.path.isdir(abs_path):
            purge_pycache(abs_path)


install_requires = parse_requirements('requirements/requirements-base.txt')
extras_require = {
    "all": parse_requirements('requirements/requirements.txt')
}
long_description = long_description_load('README.md')
purge_pycache('intelliw')

setup(
    name='intelliw',
    version=__version__,
    author='yonyou',
    author_email='yonyou@yonyou.com',
    url="http://git.yonyou.com/iuapaipaas/iw-algo-fx",
    python_requires='>=3.6',

    description='An easy to start Intelligent Workshop Algorithm Framework',
    long_description=long_description,
    long_description_content_type="text/markdown",

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
    ],

    packages=find_packages(),
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,

    entry_points={
        'console_scripts': [
            'intelliw = intelliw.interface.entrypoint:run'
        ]
    },

)



'''
Author: hexu
Date: 2021-10-28 19:43:56
LastEditTime: 2021-10-28 19:43:56
LastEditors: your name
Description: 
FilePath: /iw-algo-fx/intelliw/interface/__init__.py
'''
# from intelliw.utils import yms_downloader
from intelliw.utils.env_file_util import read_env_file
from intelliw.utils import yms_downloader_v2 as yms_downloader
# 专属化加载环境变量文件
read_env_file()
yms_downloader.run()

from intelliw.utils.logger import Logger
Logger()



import os
import sys
import inspect

# 方便别人引用, 不要删除
from intelliw.config.dataset_config import set_dataset_config


class FrameworkMode:
    """
    算法框架流程
    """
    Import = "import"
    Train = "train"
    Infer = "infer"
    Batch = "batch"
    DistTrain = "distributedtrain"


FLAME_PROF_MODE = False
FRAMEWORK_MODE = ''  # import/train/infer/batch

# SSL
INTELLIW_SSL_FILEPATH = ""

# 运行模式 SCAFFOLDING 脚手架，SERVER 服务端
RUNNING_MODE = 'SERVER'
REGISTER_CLUSTER_ADDRESS = ''  # 服务的ip➕端口  json: ['','']
# 校验镜像环境
CHECK_CERTIFIED_IMAGE = False

# basic
TENANT_ID = ''
# 实例 id，代表当前运行实例
INSTANCE_ID = ''
# 推理任务 id
INFER_ID = ''
# 任务 id，推理任务时与 INFER_ID 相同
SERVICE_ID = ''
# 是否专属化
IS_SPECIALIZATION = 0

INFER_CACHE = False
INFER_CACHE_TTL = 10

# 数据集相关
DATASET_BY_ID_ADDRESS = ""  # 通过url可以获取数据集所需的所有环境信息
SOURCE_TYPE = 0  # 0/6/7/8  # 0-空数据  6-表格数据 7-图像数据 8-文本数据
DATASET_INFO = ''
# cv: 0-自有 1-labelme 2-voc 3-coco  nlp: 20-txt 21-csv 22-json
INPUT_DATA_SOURCE_TRAIN_TYPE = 22
""" 数据集信息, json list
空 0:
表格 6:
    远程csv  1
    智能分析  2
    本地csv  3
    数据工场  5
    语义模型  9
图像 7:
    图片数据源 4
文本 8:
    nlp语料 21

CSV_PATH = ''
DATA_SOURCE_ADDRESS = ''
INPUT_ADDR = ''
INPUT_GETROW_ADDR = ''
INPUT_MODEL_ID = ''
INPUT_DATA_SOURCE_ID = ''
INPUT_DATA_SOURCE_TRAIN_TYPE = 2 cv: 0-自有 1-labelme 2-voc 3-coco  nlp: 20-txt 21-csv 22-json
"""

# 数据输出 输出数据源类型，0 空，2 智能分析, 5 数据工场
OUTPUT_DATASET_INFO = '{"sourceType":0}'
# todo 是否忽略数据集处理, 如果为 true 则训练时候，不进行数据集的下发
IGNORE_TRAIN_DATASET_FLAG = False

# 数据读取
DATA_SOURCE_READ_SIZE = 5000
DATA_SOURCE_READ_TIMEOUT = 60 # 数据读取超时时间
DATA_SOURCE_READ_LIMIT = sys.maxsize
TRAIN_DATASET_RATIO = 0.8  # 训练集比例
VALID_DATASET_RATIO = 0.2  # 验证集比例
TEST_DATASET_RATIO = 0.0  # 测试集比例
DATA_SPLIT_MODE = 1  # 数据集划分模式, -1 不分割, 0 顺序划分，1 全局随机划分，2 根据目标列随机划分
DATA_RANDOM_SEED = 1024 #数据集分割seed，当 DATA_SPLIT_MODE=1，2 时候生效

# cv数据存储文件名
CV_IMG_FILEPATH = "tmp_local_cv_image_data/"
CV_IMG_TRAIN_FILEPATH = os.path.join(CV_IMG_FILEPATH, "train/")
CV_IMG_VAL_FILEPATH = os.path.join(CV_IMG_FILEPATH, "val/")
CV_IMG_TEST_FILEPATH = os.path.join(CV_IMG_FILEPATH, "test/")
CV_IMG_ANNOTATION_FILEPATH = os.path.join(CV_IMG_FILEPATH, "annotations/")

# nlp语料存储文件名
# nlp数据格式，接口获取数据为：文件、行 或 为 本地数据（file/row/local）
NLP_CORPORA_INPUT_TYPE = 'local'
NLP_CORPORA_FILEPATH = "tmp_local_nlp_corpora_data/"
NLP_CORPORA_TRAIN_FILEPATH = os.path.join(NLP_CORPORA_FILEPATH, "train/")
NLP_CORPORA_VAL_FILEPATH = os.path.join(NLP_CORPORA_FILEPATH, "val/")
NLP_CORPORA_TEST_FILEPATH = os.path.join(NLP_CORPORA_FILEPATH, "test/")

# 大模型数据存储文件名
# 大模型数据格式，接口获取数据为：行row
LARGE_MODEL_INPUT_TYPE = 'row'
LARGE_MODEL_FILEPATH = "tmp_local_large_model_data/"
LARGE_MODEL_TRAIN_FILEPATH = os.path.join(LARGE_MODEL_FILEPATH, "train/")
LARGE_MODEL_VAL_FILEPATH = os.path.join(LARGE_MODEL_FILEPATH, "val/")
LARGE_MODEL_TEST_FILEPATH = os.path.join(LARGE_MODEL_FILEPATH, "test/")

# 推理服务
TOKEN = ''  # API 响应 token
API_EXTRAINFO = True  # API 响应包含 extra info
INFER_MULTI_PROCESS = False  # 是否多进程
INFER_MULTI_THREAD_COUNT = 0  # 自定义线程数
INFER_MULTI_PROCESS_COUNT = 0  # 自定义进程数
INFER_MAX_TASK_RATIO = 10  # 最大任务数是几倍的线程数

MAINPID = ''

# 云存储相关
STORAGE_SERVICE_PATH = ''
STORAGE_SERVICE_URL = ''
FILE_UP_TYPE = ""  # 对应的类型 AliOss/Minio

# AuthSDK
ACCESS_KEY = ''
ACCESS_SECRET = ''
GENERATOR_YHT_URL = ''
TEMPORARY_USER_COOKIE = ''

# eureka
START_EUREKA = False
EUREKA_ZONE = 'test'  # online/pre/test/daily
EUREKA_SERVER = ''  # eureka服务地址
EUREKA_APP_NAME = ''  # 注册服务名称
EUREKA_PROVIDER_ID = ''  # 注册服务租户

# 分布式
DIST_IS_MASTER = False

# Spark
SPARK_MODE = False
SPARK_MASTER_URL = ''

# checkpoint
CHECKPOINT_MODE = False
CHECKPOINT_SAVE_MAX = 100

# resources
CPU_COUNT = ""

# 是否选择基准模型
BASE_MODEL_MODE = False
BASE_MODEL_URL= ''
BASE_MODEL_LOCATION='./base_model/'

# 数据过滤条件(语义模型)
DATA_FILTER_CONDITION=''
# 热加载开关
HOT_RELOAD_MODE = False
TRAININGASSIGNMENT_MODEL_ID = ''

DATA_FACTORY_URL = ''

# 物化表项目名称
IWFACTORY_DR_PROJECT_NAME = 'ai_datafusion_er'
DATA_SOURCE_WRITE_SIZE = 3000
DATA_SOURCE_WRITE_TIMEOUT = 3600 # 数据写入超时时间

def is_server_mode():
    """
    online or local
    """
    return 'SERVER' == RUNNING_MODE


def str2bool(string: str) -> bool:
    """
    Convert a string to a boolean value.
    """
    return string.lower() == 'true'


def update_by_env():
    """
    Update module variables based on environment variable values.

    This function iterates over all variables defined in the current module (__name__)
    and checks if there is a corresponding environment variable with the same name.
    If so, it updates the variable's value to the one defined in the environment variable.

    Raises:
        KeyError: if an expected environment variable is not set.

    Returns:
        None
    """
    domain = os.environ.get('domain.url')
    useGpaas = str2bool(os.environ.get('useGpaas', 'false'))
    for k, v in globals().items():
        if not k.startswith('__') and not inspect.isfunction(v) and not inspect.ismodule(v):
            env_val = os.environ.get(k, os.environ.get(k.upper()))
            if env_val is not None and env_val != '':
                if isinstance(v, bool):
                    env_val = str2bool(env_val)
                    globals()[k] = env_val
                else:
                    try:
                        env_val = type(v)(env_val)
                        globals()[k] = env_val
                    except ValueError:
                        pass
            elif env_val == '' and isinstance(v, str):
                globals()[k] = env_val
            
            # gpaas 环境
            if useGpaas:
                if k == "access.key":
                    globals()["ACCESS_KEY"] = env_val
                if k == "access.secret":
                    globals()["ACCESS_SECRET"] = env_val
                if k == "useIPv6":
                    globals()["USE_IPV6"] = env_val
                if k == "STORAGE_SERVICE_URL":
                    globals()[k] = f"{domain}{env_val}"
                if k == "GENERATOR_YHT_URL":
                    globals()[k] = f"{domain}{env_val}"
                if k == "REPORT_ADDR":
                    globals()[k] = f"{domain}{env_val}"
                if k == "DATASET_BY_ID_ADDRESS":
                    globals()[k] = f"{domain}{env_val}"
        if useGpaas:
            # ACCESS_SECRET, ACCESS_SECRET 为空，使用 access.key, access.secret 
            if (globals().get("ACCESS_KEY") is None or len(globals().get("ACCESS_KEY"))==0) and os.environ.get("access.key") is not None \
                and (globals().get("ACCESS_SECRET") is None or len(globals().get("ACCESS_SECRET"))==0) and os.environ.get("access.secret") is not None:
                globals()["ACCESS_KEY"] = os.environ.get("access.key")
                globals()["ACCESS_SECRET"] = os.environ.get("access.secret")

            # ACCESS_SECRET, ACCESS_SECRET 为空，使用 cf_clientAccessKey, cf_clientAccessSecret
            if (globals().get("ACCESS_KEY") is None or len(globals().get("ACCESS_KEY"))==0)and os.environ.get("cf_clientAccessKey") is not None \
                and (globals().get("ACCESS_SECRET") is None or len(globals().get("ACCESS_SECRET"))==0) and os.environ.get("cf_clientAccessSecret") is not None:
                globals()["ACCESS_KEY"] = os.environ.get("cf_clientAccessKey")
                globals()["ACCESS_SECRET"] = os.environ.get("cf_clientAccessSecret")




import gc
import time
import json
import traceback
from concurrent.futures import ThreadPoolExecutor

try:
    from confluent_kafka import Producer, Consumer, KafkaError
except ImportError:
    raise ImportError("\033[31mIf use kafka, you need: pip install confluent_kafka>=2.5.3 \033[0m")
import threading
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()


class KafkaClient:

    def __init__(self, brokers: str, group_id: str = None, username: str = None, password: str = None,
                 producer_config: dict = None, consumer_config: dict = None, security_protocol: str = None,
                 sasl_mechanism: str = None, auth_type: str = None,
                 key_serializer=None, value_serializer=None, ssl_ca_location: str = None, ssl_endpoint_identification_algorithm: str = 'none'):
        """
        Kafka 通用客户端初始化
        :param brokers: Kafka 集群的地址 (例如 'localhost:9092')
        :param group_id: 消费者使用的分组ID，生产者可忽略
        :param username: SASL 验证用的用户名
        :param password: SASL 验证用的密码
        :param producer_config: 自定义的生产者配置字典
        :param consumer_config: 自定义的消费者配置字典
        :param security_protocol: 安全协议，例如 'SASL_SSL' 或 'SASL_PLAINTEXT'，若设置 auth_type 则自动推断
        :param sasl_mechanism: SASL 认证机制，'PLAIN' 或 'SCRAM-SHA-256'，若设置 auth_type 则自动推断
        :param auth_type: 认证类型 (plain, scram, scram_ssl, ssl)，会覆盖 security_protocol 和 sasl_mechanism
        :param key_serializer: 用于序列化 key 的函数
        :param value_serializer: 用于序列化 value 的函数
        :param ssl_ca_location: SSL 证书路径，若不指定则默认为 /etc/ssl/certs/ca-certificates.crt
        :param ssl_endpoint_identification_algorithm: SSL 端点标识算法，默认为 'none'，可设为 'hostname' 启用主机名验证
        """
        self.brokers = brokers
        self.group_id = group_id
        self.username = username
        self.password = password
        self.auth_type = auth_type
        self.security_protocol = security_protocol
        self.sasl_mechanism = sasl_mechanism
        self.producer_config = producer_config or {}
        self.consumer_config = consumer_config or {}
        self.key_serializer = key_serializer or self.default_serializer
        self.value_serializer = value_serializer or self.default_serializer
        self.ssl_ca_location = ssl_ca_location or '/etc/ssl/certs/ca-certificates.crt'
        self.ssl_endpoint_identification_algorithm = ssl_endpoint_identification_algorithm

        self.producer = None

        logger.info(
            f"Kafka Client initialized with brokers: {brokers}, "
            f"auth_type: {auth_type}, "
            f"security_protocol: {self.security_protocol}, "
            f"sasl_mechanism: {self.sasl_mechanism}, "
            f"producer_config: {self.producer_config}, "
            f"consumer_config: {self.consumer_config}, "
            f"ssl_ca_location: {self.ssl_ca_location}, "
            f"ssl_endpoint_identification_algorithm: {self.ssl_endpoint_identification_algorithm}")



    @staticmethod
    def default_serializer(data):
        if isinstance(data, str):
            return data.encode('utf-8')
        elif isinstance(data, (dict, list, tuple)):
            return json.dumps(data, ensure_ascii=False).encode('utf-8')
        elif isinstance(data, (int, float, bool)):
            return str(data).encode('utf-8')
        return data

    AUTH_TYPE_CONFIG = {
        'plain': {
            'security_protocol': 'SASL_PLAINTEXT',
            'sasl_mechanism': 'PLAIN'
        },
        'scram': {
            'security_protocol': 'SASL_SSL',
            'sasl_mechanism': 'SCRAM-SHA-256'
        },
        'scram_ssl': {
            'security_protocol': 'SASL_SSL',
            'sasl_mechanism': 'SCRAM-SHA-256'
        },
        'ssl': {
            'security_protocol': 'SSL',
            'sasl_mechanism': None
        },
    }

    def _get_security_config(self):
        security_protocol = self.security_protocol
        sasl_mechanism = self.sasl_mechanism

        if self.auth_type:
            auth_type_lower = self.auth_type.lower()
            if auth_type_lower in self.AUTH_TYPE_CONFIG:
                config = self.AUTH_TYPE_CONFIG[auth_type_lower]
                security_protocol = security_protocol or config['security_protocol']
                sasl_mechanism = sasl_mechanism or config['sasl_mechanism']

        ssl_config = {}
        if security_protocol and security_protocol in ('SSL', 'SASL_SSL'):
            ssl_config['ssl.ca.location'] = self.ssl_ca_location
            ssl_config['ssl.endpoint.identification.algorithm'] = self.ssl_endpoint_identification_algorithm

        if self.username and self.password:
            return {
                'sasl.mechanism': sasl_mechanism,
                'security.protocol': security_protocol,
                'sasl.username': self.username,
                'sasl.password': self.password,
                **ssl_config
            }
        elif security_protocol:
            return {
                'security.protocol': security_protocol,
                **ssl_config
            }
        return {}

    def create_producer(self):
        """创建 Kafka 生产者，优先使用用户传入配置"""
        default_producer_conf = {
            'bootstrap.servers': self.brokers,
            'linger.ms': 10,  # 批量发送时等待时间
            'batch.num.messages': 1000,  # 批量发送的最大消息数
            'queue.buffering.max.ms': 1000,  # 最大缓冲时间
            'message.timeout.ms': 30000,  # 消息超时配置
            'retries': 5,  # 消息发送失败重试次数
        }
        security_conf = self._get_security_config()
        config = {**default_producer_conf, **security_conf, **self.producer_config}
        logger.info(f"Kafka Producer Config: {config}")
        return Producer(config)

    def create_consumer(self, topics: list):
        """创建 Kafka 消费者，优先使用用户传入配置"""
        default_consumer_conf = {
            'bootstrap.servers': self.brokers,
            'group.id': self.group_id,
            'auto.offset.reset': 'earliest',  # 从最早的偏移量开始消费
            'enable.auto.commit': False,  # 手动提交偏移量
        }
        security_conf = self._get_security_config()  # 获取 SASL 安全配置
        config = {**default_consumer_conf, **security_conf, **self.consumer_config}
        logger.info(f"Kafka Consumer Config: {config}")
        consumer = Consumer(config)
        consumer.subscribe(topics)
        return consumer

    def produce_message(self, topic: str, key: str = None, value: str = None, raise_error=False):
        """发送 Kafka 消息"""
        assert not (value is None and key is None), 'Need at least one: key or value'

        serialized_key = self.key_serializer(key)
        serialized_value = self.value_serializer(value)

        assert type(serialized_key) in (bytes, bytearray, memoryview, type(None))
        assert type(serialized_value) in (bytes, bytearray, memoryview, type(None))

        if self.producer is None:
            self.producer = self.create_producer()

        try:
            self.producer.produce(topic, key=serialized_key, value=serialized_value, callback=delivery_callback)
            [self.producer.poll(0.05) for _ in range(3)]
        except Exception as e:
            logger.error(f"Error producing message to topic '{topic}', key: {serialized_key}, error: {e}")
            if raise_error:
                raise  # 直接抛出原异常，保留异常信息
        finally:
            self.producer.flush()

    def produce_messages_in_batch(self, topic: str, messages: list, poll_interval: float = 0.1):
        """
        批量发送 Kafka 消息
        :param topic: Kafka topic
        :param messages: 消息列表，包含 (key, value) 元组
        :param poll_interval: 每隔多少秒调用一次 poll() 以确保消息发送
        """
        if self.producer is None:
            self.producer = self.create_producer()

        try:
            for key, value in messages:
                self.producer.produce(topic, key=key, value=value, callback=delivery_callback)
                self.producer.poll(0)  # 定期调用以处理异步消息

            while True:
                remaining_messages = self.producer.flush(0)
                if remaining_messages == 0:
                    break
                time.sleep(poll_interval)
        except Exception as e:
            logger.error(f"Error producing message batch: {e}")

    def consume_messages(self, topics: list, handle_func, timeout=1.0, return_by_empty=False, return_by_count=0,
                         return_by_func=None):
        """
        消费 Kafka 消息
        :param topics: Kafka topic list
        :param handle_func: message 后处理方法，
            方法需要有两个参数，例如：
             - def handle_message(key, value)
             - def delivery_callback(err, msg)
            第一个参数代表consumer的 message.key(), 第二个参数为consumer的 message.value()
        :param timeout: 获取消息超时时长
        :param return_by_empty: 当获取消息 None 时结束
        :param return_by_count 当获取消息数量等于count时结束, 0为不限制
        :param return_by_func  参数传入一个方法，方法参数与handle_func相同，返回类型需要为 bool 型，当返回为 True 时，消费停止
        """

        consumer = self.create_consumer(topics)

        try:
            count = 0
            while True:
                msg = consumer.poll(timeout)
                if msg is None:
                    # 当获取消息 None 时结束
                    if return_by_empty:
                        return None
                    continue
                if msg.error():
                    if msg.error().code() != KafkaError._PARTITION_EOF:
                        logger.error(f"Kafka error: {msg.error()}")
                    continue
                handle_func(msg.key(), msg.value())  # 处理消息
                consumer.commit()  # 手动提交偏移量
                count += 1

                # 当获取消息数量等于count时结束
                if return_by_count and count >= return_by_count:
                    return None
                if return_by_func and return_by_func(msg.key(), msg.value()) is True:
                    return None

                # 定期调用垃圾回收
                if count % 10 == 0:
                    gc.collect()

        except Exception as e:
            logger.error(f"Error consuming messages: {e}, {traceback.format_exc()}")
        finally:
            consumer.close()

        return None

    def consume_messages_with_threads(self, topics: list, handle_func, num_threads=5, timeout=1.0):
        def consumer_task():
            self.consume_messages(topics, handle_func, timeout)

        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = [executor.submit(consumer_task) for _ in range(num_threads)]
            for future in futures:
                future.result()  # 等待所有线程完成任务


def delivery_callback(err, msg):
    if err:
        logger.error(f"Message failed delivery: {err}")
    else:
        logger.info(f"Message delivered to {msg.topic()} [{msg.partition()}] at offset {msg.offset()}")


def handle_message(key, value):
    logger.info(f"Received message with key: {key}, value: {value}")



import logging
import os.path
import urllib

import requests
import importlib.util
from typing import List
from intelliw.config import config
from intelliw.utils.context import header_ctx


class HijackModelLoad:
    error = type("ModelLoadError", (Exception, object), {
        "__str__": lambda self: '''算法使用了未验证的运行环境：
         1. 请使用ONNX模型进行服务推理
         2. 请勿使用Paddle，TensorFlow等框架加载模型
             ''',
        "ignore_stack": lambda self: True
    })

    def __init__(self):

        self.config: List[dict] = [
            # {'package': 'torch', 'function': 'load'},
            {'package': 'paddle', 'function': 'load'},
            {'package': 'paddle', 'function': 'jit.load'},
            {'package': 'paddle', 'function': 'Model.load'},
            {'package': 'tensorflow', 'function': 'keras.models.load_model'},
            {'package': 'tensorflow', 'function': 'saved_model.load'}
        ]
        self.hijack()

    def hijack(self):
        for c in self.config:
            package = c['package']
            function = c['function']
            if importlib.util.find_spec(package):
                p = importlib.import_module(package)
                f_list = function.split(".")
                for f in f_list:
                    try:
                        setattr(p, f, HijackModelLoad._raise)
                        p = getattr(p, f)
                    except Exception as e:
                        print('hijack error:', e)

    def _raise(*args, **kwargs):
        raise HijackModelLoad.error


class HijackRequests:
    """
    注入requests包，在专属化环境写入证书信息，通过工作坊给的链接进行下载，然后将文件夹位置赋值给环境变量INTELLIW_SSL_FILEPATH
    如果都验证没问题，所有请求都注入证书, 否则都避开证书校验
    """

    @staticmethod
    def request(method, url, **kwargs):
        keywords = header_ctx.get()
        if 'headers' in kwargs:
            kwargs['headers'].update(keywords)
        else:
            kwargs['headers'] = keywords

        if config.INTELLIW_SSL_FILEPATH:
            path = config.INTELLIW_SSL_FILEPATH
            # 根据host判断的方式
            if os.path.isdir(path):
                var = urllib.parse.urlparse(url).netloc
                file_paths = [os.path.join(path, file_name) for file_name in os.listdir(path) if
                              os.path.splitext(file_name)[0] == var]
                if file_paths:
                    kwargs['verify'] = file_paths[0]
        else:
            kwargs['verify'] = False
        with requests.sessions.Session() as session:
            try:
                return session.request(method=method, url=url, **kwargs)
            except requests.exceptions.SSLError:
                kwargs['verify'] = False
                return session.request(method=method, url=url, **kwargs)

    def __init__(self):
        self.hijack()

    @staticmethod
    def hijack():
        requests.api.request = HijackRequests.request
        requests.request = HijackRequests.request


class HijackLogging:
    def __init__(self):
        self.hijack()

    @staticmethod
    def hijack():
        logging.LogRecord.traceid = '0'




import json
import time

from intelliw.utils.kafka import get_client, handle_message
import os

hosts = '192.168.15.218:9092'

# producer_config={'acks': 'all'},
consumer_config={'auto.offset.reset': 'earliest',
                         'enable.auto.commit': True,
                         'max.poll.interval.ms': 8000000,
                         'heartbeat.interval.ms': 3000,
                         'session.timeout.ms': 30000}
security_protocol: str = 'SASL_PLAINTEXT'
sasl_mechanism: str = 'PLAIN'
def test_kafka():
    kafka_client = get_client(brokers=hosts, 
                              group_id='yy-test-group-id',
                              consumer_config=consumer_config,
                              security_protocol=security_protocol,
                              sasl_mechanism=sasl_mechanism,
                              )
    kafka_client.produce_message(topic='test-topic', value={'nihao': '1'})
    kafka_client.consume_messages(topics=['test-topic'], handle_func=handle_message)
test_kafka()


'''
Author: hexu
Date: 2021-11-19 10:29:02
LastEditTime: 2023-05-23 18:48:50
LastEditors: Hexu
Description: 文件存储服务
FilePath: /iw-algo-fx/intelliw/utils/storage_service.py
'''
import json
import os
import io
import mimetypes
import traceback
import logging
from intelliw.utils import iuap_request
from intelliw.config import config
import urllib.parse as urlparse


class StorageService:
    """
    文件传输，通过服务端获取的临时url进行文件在多个云存储（AliOss/Minio/HWObs）上的操作
    所有的请求需要进行加签
    """

    def __init__(self, key, process_type):
        """
        初始化下载/上传链接
            Args:
                key :
                    upload ： 文件存储于桶的位置（包含文件名） OSS_PATH + filename
                    download : AI工作坊存储在环境变量的值     DATA_SOURCE_ADDRESS
                process_type : download/upload
            Returns:
                self.service_url 为服务端根据client_type，process_type生成的链接， download和upload操作需要此链接
        """
        self.key = key
        self.service_url = self.__client_init(
            key, process_type)
        self.content_type = mimetypes.guess_type(
            key)[0] or 'application/octet-stream'
        if self.content_type == "text/csv":
            self.content_type = 'application/octet-stream'

    def __client_init(self, key, process_type):
        """
        从服务端获取云存储操作链接

        下载链接一般可以直接下载的
        """
        data = {
            'key': key,
            'UrlType': process_type,
            'instanceid': config.INSTANCE_ID
        }
        client_type = self._get_client_type(config.FILE_UP_TYPE)
        if client_type != "":
            data["clientType"] = client_type

        # shit
        if os.environ.get('useGpaas') == 'true':
            STORAGE_SERVICE_URL = f"{os.environ.get('domain.url')}{os.environ.get('STORAGE_SERVICE_URL')}" or \
                                  config.STORAGE_SERVICE_URL
        else:
            STORAGE_SERVICE_URL = os.environ.get('STORAGE_SERVICE_URL') or config.STORAGE_SERVICE_URL
        if not STORAGE_SERVICE_URL:
            raise EnvironmentError("STORAGE_SERVICE_URL is not set")
        resp = iuap_request.get(url=STORAGE_SERVICE_URL, params=data)
        resp.raise_for_status()
        result = resp.json
        assert result.get("status") == 1, result
        return result.get("data")

    def upload(self, file):
        """
        通过操作链接进行上传操作
            Args:
                file ： 上传文件本地路径/文件流
        """

        def _put_file(f):
            headers = {'Content-Type': self.content_type}
            resp = iuap_request.put_file(
                url=self.service_url, headers=headers, data=f)
            resp.raise_for_status()

        if isinstance(file, str) and os.path.exists(file) and os.path.isfile(file):
            with open(file, 'rb') as f:
                _put_file(f)
        elif isinstance(file, (bytes, bytearray, io.BufferedIOBase)):
            _put_file(file)
        else:
            file = json.dumps(file, ensure_ascii=False).encode('utf-8')
            _put_file(file)

    def download(self, output_path, stream=False):
        """
        通过操作链接进行下载操作
            Args:
                output_path ： 下载文件保存路径
        """
        if os.path.exists(output_path):
            raise FileExistsError(f'文件 {output_path} 已存在')
        if stream:
            iuap_request.stream_download(
                method="get", url=self.service_url, output_path=output_path)
        else:
            iuap_request.download(url=self.service_url,
                                  output_path=output_path)

    def _get_client_type(self, _type):
        env_val = _type.upper()
        if env_val == "MINIO":
            return "Minio"
        elif env_val == "ALIOSS":
            return "AliOss"
        elif env_val == "HWOBS":
            return "HWObs"
        else:
            return ""


class FileTransferDevice(Exception):
    """FileTransferDevice 文件传输引擎

    使用方法:
        >>> from intelliw.feature import FileTransferDevice
        >>> ftd = FileTransferDevice(fileobj/filepath, "report_type", filename="xxx.csv")

    如果文件上传后需要中断服务（e.g. 数据集检查出错，上传检查报告后，中断训练服务）
        >>> if "存在某些错误":
        >>>     ftd.msg = "xxxxx错误"
        >>>     raise ftd

    :param file 必须 上传的文件, 可以是路径, 也可以是open后的文件
    :param transfer_type 必须 上传文件的业务类型
    :param filename 非必需 上传后文件的名称，类型需要与源文件一致(.csv/.json)
    :param msg 非必需 错误信息
    """

    def __init__(self, file, transfer_type, filename="data", msg="") -> None:
        self.url = ""
        self.curkey = None
        self.msg = msg
        self.upload_msg = None
        self.transfer_type = transfer_type
        self.__put_file__(file, filename)

    def __put_file__(self, file, filename) -> None:
        import intelliw.utils.message as message
        from intelliw.utils.global_val import gl

        try:
            if config.is_server_mode():
                # filename
                if isinstance(file, str) and os.path.isfile(file):
                    filename = os.path.basename(file)

                # s3 path
                self.curkey = os.path.join(
                    config.STORAGE_SERVICE_PATH, config.SERVICE_ID, filename
                )

                # file upload
                uploader = StorageService(self.curkey, "upload")
                uploader.upload(file)

                url = urlparse.urlparse(uploader.service_url)
                self.url = url.scheme + "://" + url.netloc + url.path

                # report
                if gl.recorder:
                    gl.recorder.report(message.CommonResponse(
                        200, 'file_transfer', 'success',
                        {'filepath': self.curkey},
                        businessType=self.transfer_type))
        except Exception as e:
            self.upload_msg = e
            logging.error("上传文件错误: %s, %s", e, traceback.format_exc())

    def __str__(self) -> str:
        msg = f"算法服务中断: {self.msg}\n中断类型: {self.transfer_type}\n"
        if self.upload_msg is None:
            msg += f'文件上传：成功\n中断原因查看：{self.curkey}'
        else:
            msg = f'文件上传：失败\n上传失败信息: {self.upload_msg}'
        return msg

    def ignore_stack(self):
        """
        error ignore stack
        """
        return True

    def btype(self):
        """
        error btype
        """
        return self.transfer_type



'''
Author: Hexu
Date: 2022-08-16 10:18:29
LastEditors: Hexu
LastEditTime: 2023-05-10 14:21:04
FilePath: /iw-algo-fx/intelliw/utils/spark_process/engine.py
Description: data process engine
'''
import inspect
import threading
from functools import wraps
from typing import Tuple, Union
import numpy as np
import pandas as pd
from random import Random
from intelliw.utils.exception import DatasetException, FeatureProcessException
from intelliw.config.config import SPARK_MODE, SPARK_MASTER_URL, FRAMEWORK_MODE, FrameworkMode
from intelliw.utils.logger import _get_framework_logger

logger = _get_framework_logger()

# 只有训练才需要spark
if SPARK_MODE and FRAMEWORK_MODE in (FrameworkMode.DistTrain, FrameworkMode.Train):
    try:
        from intelliw.utils.spark_process.spark import Spark
        from pyspark import pandas as ps
        from pyspark.ml.feature import (Bucketizer, MaxAbsScaler, MinMaxScaler,
                                        QuantileDiscretizer, StandardScaler,
                                        VectorAssembler)
        from pyspark.pandas.series import Series as SparkSeries
        from pyspark.sql import functions as F
        from pyspark.ml.functions import vector_to_array
        from pyspark.sql.types import DoubleType, IntegerType
        from pyspark.sql.dataframe import DataFrame as SparkDF

        _unlist = F.udf(lambda x: float(list(x).pop()), DoubleType())
    except ImportError:
        raise ImportError(
            "\033[31mIf use spark, you need: pip install pyspark and pyarrow>=2.0.0\033[0m")
else:
    SPARK_MODE = False

random = Random(1024)


class CsvProcessEngine:
    type = "csv"

    def __init__(self, filepath: str, local_mode=True) -> None:
        self.filepath = filepath
        if local_mode:
            self.spark_mode = False
        else:
            self.spark_mode = SPARK_MODE

        if self.spark_mode:
            self.spark_core = Spark.get_spark(master=SPARK_MASTER_URL)

    def read(self, nrows=None, skiprows=None) -> list:
        if self.spark_mode:
            csv = self.spark_core.read_file(
                CsvProcessEngine.type, self.filepath)
            if skiprows is not None:
                csv = csv.rdd.zipWithIndex().filter(
                    lambda x: x[1] > skiprows).map(lambda x: x[0]).toDF()
            if nrows is not None:
                csv = csv.limit(nrows)
            return csv.toPandas()
        else:
            coding = ["utf8", "GB2312", "GBK", "GB18030", "utf-8-sig"]
            for c in coding:
                try:
                    return pd.read_csv(self.filepath, nrows=nrows, skiprows=skiprows, encoding=c)
                except UnicodeDecodeError:
                    continue
                except pd.errors.EmptyDataError:
                    raise DatasetException("empty csv")
            else:
                raise UnicodeDecodeError(
                    f"Can Not Encoding, Support Coding: {coding}")


class Engine:
    _instance_lock = threading.Lock()
    spark_mode = SPARK_MODE

    def __new__(cls):
        if not hasattr(Engine, "_instance"):
            with Engine._instance_lock:
                if not hasattr(Engine, "_instance"):
                    if Engine.spark_mode:
                        Engine.spark_core = Spark.get_spark(master=SPARK_MASTER_URL)
                    Engine._instance = object.__new__(cls)
                    _mode = "Spark" if Engine.spark_mode else "Normal"
                    logger.info(
                        f"\033[33mData Process Engine: {_mode}\033[0m")
        return Engine._instance

    def __init__(self) -> None:
        self.pd = pd
        if Engine.spark_mode:
            self.ps = Engine.spark_core.ps
            self.ps.set_option('compute.ops_on_diff_frames', True)
            SparkFeatureProcess._hijacked()
        else:
            self.ps = None

    def isna(self, obj):
        if Engine.spark_mode:
            return self.ps.isna(obj)
        return self.pd.isna(obj)

    def csv(self, filepath) -> CsvProcessEngine:
        return CsvProcessEngine(filepath)

    def DataFrame(self, data=None, index=None, columns=None, dtype=None, copy=False):
        if Engine.spark_mode:
            df = self.ps.DataFrame(
                data, index=index, columns=columns, dtype=dtype, copy=copy)
        else:
            df = self.pd.DataFrame(data, index=index,
                                   columns=columns, dtype=dtype, copy=copy)
        return df

    def concat(self, objs: list, axis: int = 0, join: str = "outer", ignore_index: bool = False, sort: bool = False):
        if Engine.spark_mode:
            df = self.ps.concat(objs, axis=axis, join=join,
                                ignore_index=ignore_index, sort=sort)
        else:
            df = self.pd.concat(objs, axis=axis, join=join,
                                ignore_index=ignore_index, sort=sort)
        return df

    def to_numeric(self, arg, errors='raise'):
        if Engine.spark_mode:
            df = self.ps.to_numeric(arg, errors)
        else:
            df = self.pd.to_numeric(arg, errors)
        return df

    def cut(self, x,
            bins,
            right: bool = True,
            labels=None,
            retbins: bool = False,
            precision: int = 3,
            include_lowest: bool = False,
            duplicates: str = "raise"):
        if Engine.spark_mode:
            x, _ = SparkFeatureProcess.bucket({'boundaries': bins}, x, "")
        else:
            df = self.pd.cut(x, bins, right, labels, retbins,
                             precision, include_lowest, duplicates)
        return df

    def pivot_table(self, data,
                    values=None,
                    index=None,
                    columns=None,
                    fill_value=None, ):
        if Engine.spark_mode:
            df = data.to_spark()
            df = df.groupBy(index).pivot(columns).sum(values)
            df = df.pandas_api()
        else:
            df = self.pd.pivot_table(
                data, values=values, index=index, columns=columns, fill_value=fill_value)
        return df


class SparkFeatureProcess:

    @staticmethod
    def _hijacked():
        from intelliw.functions import feature_process as fp
        for f in dir(SparkFeatureProcess):
            if not f.startswith("_") \
                    and hasattr(fp, f) \
                    and inspect.isfunction(getattr(fp, f)):
                hijacked_func = getattr(SparkFeatureProcess, f)
                setattr(fp, f, hijacked_func)

    # feature_process.
    def _dataframe_process(f):
        @wraps(f)
        def inner(params, col_frame, mode):
            # input
            if isinstance(col_frame, SparkSeries):
                col_frame = col_frame.to_frame()
            col_frame = col_frame.to_spark()

            # do process
            result = f(params, col_frame, mode)

            # output
            is_multi_result, frame_idx = type(result) == tuple, None
            if is_multi_result:
                result = list(result)
                for idx, c in enumerate(result):
                    if isinstance(c, SparkDF):
                        col_frame, frame_idx = c.pandas_api(), idx
                        break
            elif isinstance(result, SparkDF):
                col_frame = result.pandas_api()
            else:
                return result

            if col_frame.shape[-1] == 1:
                col_frame = col_frame[col_frame.columns[0]]
            if is_multi_result and frame_idx is not None:
                result[frame_idx] = col_frame
                return result
            else:
                return col_frame

        return inner

    @staticmethod
    def _to_vector(df: object) -> object:
        col_name, new_col = df.columns.pop(), "new_col"
        assembler = VectorAssembler(
            inputCols=[col_name], outputCol=new_col, handleInvalid="keep")
        df: object = assembler.transform(df)
        df = df.select(new_col).withColumnRenamed(new_col, col_name)
        return df

    @staticmethod
    @_dataframe_process
    def ordinal(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[object, Union[str, dict, list]]:
        col_name = col_frame.columns.pop()
        category = col_frame.select(col_name).distinct().collect()
        category_list = [i[col_name] for i in category]
        params['category_list'] = category_list

        if col_frame.where(~col_frame[col_name].isin(category_list)).count() > 0:
            # when not in category_list is len(category_list) other keep the data
            col_frame = col_frame.withColumn(
                col_name,
                F.when(
                    ~F.col(col_name).isin(
                        category_list), len(category_list)
                ).otherwise(
                    F.col(col_name)
                ))

        # ordinal
        for i, cate in enumerate(category_list):
            col_frame = col_frame.withColumn(
                col_name,
                F.when(
                    F.col(col_name) == cate, int(i)
                ).otherwise(
                    F.col(col_name)
                ))
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def one_hot(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[object, Union[str, dict, list]]:
        col_name, drop_col_name = col_frame.columns.pop(), "drop_col"
        category = col_frame.select(col_name).distinct().collect()
        category_list = [i[col_name] for i in category]
        params['category_list'] = category_list
        col_frame = col_frame.withColumnRenamed(col_name, drop_col_name)
        for i, cate in enumerate(category_list + [None]):
            col_frame = col_frame.withColumn(
                str(i),
                F.when(
                    F.col(drop_col_name) == cate, 1
                ).otherwise(
                    0
                ))
        col_frame = col_frame.drop(drop_col_name)
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def quantile_transform(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name, new_col_name = col_frame.columns.pop(), "new_col"

        bucket_size = max(2, params.get('bucket_size', 10))
        model = QuantileDiscretizer(numBuckets=bucket_size, inputCol=col_name,
                                    outputCol=new_col_name, relativeError=0.01,
                                    handleInvalid="keep").fit(col_frame)
        boundaries = model.getSplits()
        params['boundaries'] = boundaries[1:-1]  # drop -inf inf
        col_frame = model.transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name, F.col(new_col_name).cast(IntegerType())
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def log_transform(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name = col_frame.columns.pop()
        if col_frame.where((col_frame[col_name].isNotNull()) & (col_frame[col_name] < 0)).count() > 0:
            raise FeatureProcessException(
                'log_transform transform can only handle the data bigger than zero! while there are negetive data in dataframe')

        col_frame = col_frame.withColumn(
            col_name,
            F.when(
                F.col(col_name).isNotNull(),
                F.log(float(10), F.col(col_name) + 1)
            ).otherwise(
                F.col(col_name)
            ))
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def standard_scale(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name, new_col_name = col_frame.columns.pop(), "new_col"
        col_frame = SparkFeatureProcess._to_vector(col_frame)
        model = StandardScaler(
            inputCol=col_name, outputCol=new_col_name,
            withStd=True, withMean=True
        ).fit(col_frame)
        params['mean'] = list(model.mean).pop()
        params['std'] = list(model.std).pop()

        col_frame = model.transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name,
            _unlist(new_col_name)
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def max_abs_scale(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name, new_col_name = col_frame.columns.pop(), "new_col"
        col_frame = SparkFeatureProcess._to_vector(col_frame)
        model = MaxAbsScaler(
            inputCol=col_name, outputCol=new_col_name
        ).fit(col_frame)
        params['max_abs'] = list(model.maxAbs).pop()

        col_frame = model.transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name,
            _unlist(new_col_name)
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def min_max_scale(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name, new_col_name = col_frame.columns.pop(), "new_col"
        col_frame = SparkFeatureProcess._to_vector(col_frame)
        model = MinMaxScaler(
            inputCol=col_name, outputCol=new_col_name
        ).fit(col_frame)
        params['max'] = list(model.originalMax).pop()
        params['min'] = list(model.originalMin).pop()
        col_frame = model.transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name,
            _unlist(new_col_name)
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def bucket(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[object, Union[str, dict, list]]:
        if not ('boundaries' in params and len(params['boundaries']) > 0):
            raise FeatureProcessException(
                "for bucket transform ,boundaries can't be empty!")
        col_name, new_col_name = col_frame.columns.pop(), "new_col"
        bins = [float('-inf')] + params['boundaries'] + [float('inf')]

        model = Bucketizer(
            splits=bins, inputCol=col_name, outputCol=new_col_name, handleInvalid="keep")
        col_frame = model.transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name, F.col(new_col_name).cast(IntegerType())
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def translation(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name = col_frame.columns.pop()
        min = col_frame.rdd.min()[col_name]
        params['min'] = min
        if min < 0:
            col_frame = col_frame.withColumn(
                col_name, F.col(col_name) + abs(min))
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def square(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[object, Union[str, dict, list]]:
        col_name = col_frame.columns.pop()
        col_frame = col_frame.withColumn(
            col_name,
            F.when(
                col_frame[col_name].isNotNull(), F.col(col_name) *
                                                 F.col(col_name)
            ).otherwise(
                F.col(col_name)
            ))
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def robust_scale(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        col_name = col_frame.columns.pop()
        median, quant25, quant75 = col_frame.approxQuantile(
            col_name, [0.5, 0.25, 0.75], 0.01)
        params['median'] = median
        params['quant25'] = quant25
        params['quant75'] = quant75
        if quant25 != quant75:
            col_frame = col_frame.withColumn(
                col_name,
                F.when(
                    F.col(col_name).isNotNull(),
                    (F.col(col_name) - median) / (quant75 - quant25)
                ).otherwise(
                    F.col(col_name)
                ))
        return col_frame, params

    @staticmethod
    def drop(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[object, Union[str, dict, list]]:
        return ps.DataFrame(), params

    @staticmethod
    @_dataframe_process
    def hash_feature(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        hash_bucket_size = max(2, params.get('bucket_size', 10))
        params['bucket_size'] = hash_bucket_size
        col_name = col_frame.columns.pop()
        tohash = F.udf(lambda x: hash(x) % hash_bucket_size, IntegerType())
        col_frame = col_frame.withColumn(
            col_name,
            F.when(
                F.col(col_name).isNotNull(),
                tohash(col_name)
            ).otherwise(
                F.col(col_name)
            ))
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def interval_bucket(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        bucket_size = params.get('bucket_size', 10)
        boundaries = []

        col_name, new_col_name = col_frame.columns.pop(), "new_col"
        _rdd = col_frame.rdd
        minim = _rdd.min()[col_name]
        maxim = _rdd.max()[col_name]
        del _rdd

        for pos in range(1, bucket_size):
            boundaries.append((maxim - minim) * pos / bucket_size + minim)
        params['boundaries'] = boundaries

        bins = [float('-inf')]
        bins.extend(boundaries)
        bins.append(float('inf'))

        bucketizer = Bucketizer(
            splits=bins, inputCol=col_name, outputCol=new_col_name)
        col_frame = bucketizer.setHandleInvalid("keep").transform(col_frame)
        col_frame = col_frame.select(
            new_col_name
        ).withColumn(
            new_col_name,
            F.col(new_col_name).cast(IntegerType())
        ).withColumnRenamed(
            new_col_name, col_name
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def kmeans_bucket(params: Union[str, dict, list], col_frame: object, mode: str) -> Tuple[
        object, Union[str, dict, list]]:
        from sklearn.cluster import KMeans
        col_name = col_frame.columns.pop()
        data = col_frame.pandas_api().to_numpy().reshape(-1, 1)
        k = max(2, params.get('bucket_size', 10))
        kmeans = KMeans(n_clusters=k).fit(data)
        centers = kmeans.cluster_centers_
        params['cluster_centers'] = centers
        dist = [abs(data - c).reshape(-1) for c in centers]
        label = np.array(dist).argmin(axis=0)
        col_frame = Spark.sc.parallelize(
            label
        ).map(
            lambda x: (int(x) if x is not None else None,)
        ).toDF(
            [col_name]
        )
        return col_frame, params

    @staticmethod
    @_dataframe_process
    def numeric_miss_check_repair(check_repair_config: Union[str, dict, list], col_frame: object, mode: str) -> object:
        col_name = col_frame.columns.pop()
        group_cols = check_repair_config['group_cols']
        index = check_repair_config['index']
        mr = check_repair_config['miss_repair']

        if mr == '1':
            mean = col_frame.pandas_api().mean()[0]
            col_frame = col_frame.fillna(mean)
        elif mr == '2':
            col_frame = col_frame.fillna(0)
        elif mr == '3':
            median = col_frame.approxQuantile(col_name, (0.5,), 0.01)[0]
            col_frame = col_frame.fillna(median)
        elif mr == '4':
            mode = col_frame.groupby(col_name).count().orderBy(
                "count", ascending=False).first()[0]
            col_frame = col_frame.fillna(mode)
        elif mr == '6':
            pass
        elif type(check_repair_config['miss_repair']) == dict:
            if check_repair_config['miss_repair']['repair_type'] == '7':
                if not ('quantile_low' in check_repair_config['miss_repair'] and 'quantile_high' in check_repair_config[
                    'miss_repair']):
                    raise FeatureProcessException(
                        f"origin_col_num:{index}, when miss_repair='7', quantile_low and  quantile_high must be set")
                low = check_repair_config['miss_repair']['quantile_low']
                high = check_repair_config['miss_repair']['quantile_high']
                if not (0.0 <= low <= high <= 1.0):
                    raise FeatureProcessException(
                        f"origin_col_num:{index}, quantile_low and quantile_high must between 0.0 and 1.0, and quantile_low must be smaller than quantile_high ")

                quantile_high = col_frame.approxQuantile(
                    col_name, (high,), 0.01)[0]
                quantile_low = col_frame.approxQuantile(
                    col_name, (low,), 0.01)[0]
                val = quantile_low + \
                      (quantile_high - quantile_low) * random.random()

                if check_repair_config['type'].lower() in ['int', 'integer']:
                    val = int(val)
                col_frame = col_frame.fillna(val)
        return col_frame

    @staticmethod
    @_dataframe_process
    def numeric_outlier_check(check_repair_config, col_frame, mode):
        index = check_repair_config['index']
        if 'outlier_check' not in check_repair_config:
            raise FeatureProcessException(
                f"origin_col_num:{index}, outlier_check config missing!")
        if not (check_repair_config['outlier_check'] in ['1', '2', '3']):
            raise FeatureProcessException(
                f"origin_col_num:{index}, outlier_check must be in ['1', '2', '3']!")

        low, high = 0, 0
        col_name = col_frame.columns.pop()
        # 3 西格玛
        if check_repair_config['outlier_check'] == '1':
            new_col_name = "new_col"
            col_frame = SparkFeatureProcess._to_vector(col_frame)
            model = StandardScaler(
                inputCol=col_name, outputCol=new_col_name,
                withStd=True, withMean=True
            ).fit(col_frame)
            mean = list(model.mean).pop()
            std = list(model.std).pop()
            low = mean - 3 * std
            high = mean + 3 * std
        elif check_repair_config['outlier_check'] == '2':
            quantile25 = col_frame.approxQuantile(col_name, (0.25,), 0.01)[0]
            quantile75 = col_frame.approxQuantile(col_name, (0.75,), 0.01)[0]
            low = quantile25 - 1.5 * (quantile75 - quantile25)
            high = quantile75 + 1.5 * (quantile75 - quantile25)
        return low, high

    @staticmethod
    @_dataframe_process
    def numeric_outlier_repair(check_repair_config, col_frame, mode):
        index = check_repair_config['index']
        if not ('outlier_repair' in check_repair_config):
            raise FeatureProcessException(
                f"origin_col_num:{index}, outlier_repair config missing!")
        if not (check_repair_config['outlier_repair'] in ['1', '2', '3']):
            raise FeatureProcessException(
                f"origin_col_num:{index}, outlier_repair must in ['1', '2', '3']!")

        col_name = col_frame.columns.pop()
        if check_repair_config['outlier_repair'] == '1':
            new_col_name = "new_col"
            col_frame = SparkFeatureProcess._to_vector(col_frame)
            model = StandardScaler(
                inputCol=col_name, outputCol=new_col_name,
                withStd=True, withMean=True
            ).fit(col_frame)
            mean = list(model.mean).pop()
            std = list(model.std).pop()
            low = mean - 3 * std
            high = mean + 3 * std
            col_frame = col_frame.select(
                vector_to_array(col_name).alias(col_name))
            col_frame = col_frame.withColumn(
                col_name,
                F.when(
                    F.col(col_name)[0] < low, low
                ).when(
                    F.col(col_name)[0] > high, high
                ).otherwise(
                    F.col(col_name)[0]
                ))
        elif check_repair_config['outlier_repair'] == '2':
            quantile25 = col_frame.approxQuantile(col_name, (0.25,), 0.01)[0]
            quantile75 = col_frame.approxQuantile(col_name, (0.75,), 0.01)[0]
            low = quantile25 - 1.5 * (quantile75 - quantile25)
            high = quantile75 + 1.5 * (quantile75 - quantile25)

            col_frame = col_frame.withColumn(
                col_name,
                F.when(
                    F.col(col_name) < low, low
                ).when(
                    F.col(col_name) > high, high
                ).otherwise(
                    F.col(col_name)
                ))
        return col_frame

    @staticmethod
    @_dataframe_process
    def category_miss_check_repair(check_repair_config, col_frame, mode):
        col_name = col_frame.columns.pop()
        if check_repair_config['miss_repair'] == '4':
            mode = col_frame.groupby(col_name).count().orderBy(
                "count", ascending=False).first()[0]
            col_frame = col_frame.fillna(mode)
        elif check_repair_config['miss_repair'] == '6':
            pass
        return col_frame
